from typing import Type, Mapping, Any, Iterable, TypeVar, Tuple, Callable, Generator

from datapipelines import DataSource, DataSink, PipelineContext, validate_query, NotFoundError
from merakicommons.cache import Cache as CommonsCache

from . import uniquekeys
from ..core.championmastery import ChampionMastery
from ..core.league import Leagues, ChallengerLeague, MasterLeague
from ..core.staticdata import Champion, Mastery, Rune, Item, SummonerSpell, Map, Realms, ProfileIcon, Locales, LanguageStrings, Versions, SummonerSpells, Items, Champions, Masteries, Runes, Maps, ProfileIcons
from ..core.masterypage import MasteryPage
from ..core.match import Match
from ..core.runepage import RunePage
from ..core.summoner import Summoner
from ..core.status import ShardStatus
from ..core.spectator import CurrentMatch, FeaturedMatches


T = TypeVar("T")


class Cache(DataSource, DataSink):
    def __init__(self, expiration: Mapping[type, float] = None) -> None:
        self._cache = CommonsCache()
        self._expiration = dict(expiration) if expiration is not None else {}

    @DataSource.dispatch
    def get(self, type: Type[T], query: Mapping[str, Any], context: PipelineContext = None) -> T:
        pass

    @DataSource.dispatch
    def get_many(self, type: Type[T], query: Mapping[str, Any], context: PipelineContext = None) -> Iterable[T]:
        pass

    @DataSink.dispatch
    def put(self, type: Type[T], item: T, context: PipelineContext = None) -> None:
        pass

    @DataSink.dispatch
    def put_many(self, type: Type[T], items: Iterable[T], context: PipelineContext = None) -> None:
        pass

    def _get(self, type: Any, query: Mapping[str, Any], key_function: Callable[[Mapping[str, Any]], Any], context: PipelineContext = None):
        key = key_function(query)
        try:
            return self._cache.get(type, key)
        except KeyError as e:
            raise NotFoundError from e

    def _get_many(self, type: Any, query: Mapping[str, Any], key_generator: Callable[[Mapping[str, Any]], Any], context: PipelineContext = None):
        for key in key_generator(query):
            try:
                yield self._cache.get(type, key)
            except KeyError as e:
                raise NotFoundError from e

    @staticmethod
    def _put_many_generator(items: Iterable[T], key_function: Callable[[T], Any], *key_function_arg_lists: tuple) -> Generator[Tuple[Any, T], None, None]:
        for item in items:
            for key_function_args in key_function_arg_lists:
                yield key_function(item, *key_function_args), item

    def _put(self, type: Type[T], item: T, key_function: Callable[[T], Any], *key_function_arg_lists: tuple, context: PipelineContext = None) -> None:
        try:
            expire_seconds = self._expiration[type]
        except KeyError:
            expire_seconds = -1.0

        if not key_function_arg_lists:
            key = key_function(item)
            self._cache.put(type, key, item, expire_seconds)
        else:
            for key_function_args in key_function_arg_lists:
                key = key_function(item, *key_function_args)
                self._cache.put(type, key, item, expire_seconds)
        # TODO: Put EXPIRATION into context once cache expiration works

    def _put_many(self, type: Type[T], items: Iterable[T], key_function: Callable[[T], Any], *key_function_arg_lists: tuple, context: PipelineContext = None) -> None:
        try:
            expire_seconds = self._expiration[type]
        except KeyError:
            expire_seconds = -1.0

        for key, item in Cache._put_many_generator(items, key_function, *key_function_arg_lists):
            self._cache.put(type, key, item, expire_seconds)
        # TODO: Put EXPIRATION into context once cache expiration works

    ########################
    # Champion Mastery API #
    ########################

    @get.register(ChampionMastery)
    @validate_query(uniquekeys.validate_champion_mastery_query, uniquekeys.convert_region_to_platform)
    def get_champion_mastery(self, query: Mapping[str, Any], context: PipelineContext = None) -> Champion:
        return self._get(ChampionMastery, query, uniquekeys.for_champion_mastery_query, context)

    @get_many.register(ChampionMastery)
    @validate_query(uniquekeys.validate_many_champion_mastery_query, uniquekeys.convert_region_to_platform)
    def get_many_champion_mastery(self, query: Mapping[str, Any], context: PipelineContext = None) -> Generator[Champion, None, None]:
        return self._get_many(ChampionMastery, query, uniquekeys.for_many_champion_mastery_query, context)

    @put.register(ChampionMastery)
    def put_champion_mastery(self, item: ChampionMastery, context: PipelineContext = None) -> None:
        self._put(ChampionMastery, item, uniquekeys.for_champion_mastery,
                  ("id", "id"),
                  ("id", "name"),
                  ("account.id", "id"),
                  ("account.id", "name"),
                  ("name", "id"),
                  ("name", "name"),
                  context=context)

    @put_many.register(ChampionMastery)
    def put_many_champion_mastery(self, items: Iterable[ChampionMastery], context: PipelineContext = None) -> None:
        self._put_many(ChampionMastery, items, uniquekeys.for_champion_mastery,
                       ("id", "id"),
                       ("id", "name"),
                       ("account.id", "id"),
                       ("account.id", "name"),
                       ("name", "id"),
                       ("name", "name"),
                       context=context)

    ##############
    # League API #
    ##############

    @get.register(Leagues)
    @validate_query(uniquekeys.validate_leagues_query, uniquekeys.convert_region_to_platform)
    def get_league_summoner(self, query: Mapping[str, Any], context: PipelineContext = None) -> Leagues:
        return self._get(Leagues, query, uniquekeys.for_leagues_query, context)

    @get_many.register(Leagues)
    @validate_query(uniquekeys.validate_many_leagues_query, uniquekeys.convert_region_to_platform)
    def get_many_league_summoner(self, query: Mapping[str, Any], context: PipelineContext = None) -> Generator[Leagues, None, None]:
        return self._get_many(Leagues, query, uniquekeys.for_many_leagues_query, context)

    @put.register(Leagues)
    def put_league_summoner(self, item: Leagues, context: PipelineContext = None) -> None:
        self._put(Leagues, item, uniquekeys.for_leagues, context=context)

    @put_many.register(Leagues)
    def put_many_league_summoner(self, items: Iterable[Leagues], context: PipelineContext = None) -> None:
        self._put_many(Leagues, items, uniquekeys.for_leagues, context=context)

    # Challenger

    @get.register(ChallengerLeague)
    @validate_query(uniquekeys.validate_challenger_league_query, uniquekeys.convert_region_to_platform)
    def get_league_summoner(self, query: Mapping[str, Any], context: PipelineContext = None) -> ChallengerLeague:
        return self._get(ChallengerLeague, query, uniquekeys.for_challenger_league_query, context)

    @get_many.register(ChallengerLeague)
    @validate_query(uniquekeys.validate_many_challenger_league_query, uniquekeys.convert_region_to_platform)
    def get_many_league_summoner(self, query: Mapping[str, Any], context: PipelineContext = None) -> Generator[ChallengerLeague, None, None]:
        return self._get_many(ChallengerLeague, query, uniquekeys.for_many_challenger_league_query, context)

    @put.register(ChallengerLeague)
    def put_league_summoner(self, item: ChallengerLeague, context: PipelineContext = None) -> None:
        self._put(ChallengerLeague, item, uniquekeys.for_challenger_league, context=context)

    @put_many.register(ChallengerLeague)
    def put_many_league_summoner(self, items: Iterable[ChallengerLeague], context: PipelineContext = None) -> None:
        self._put_many(ChallengerLeague, items, uniquekeys.for_challenger_league, context=context)

    # Master

    @get.register(MasterLeague)
    @validate_query(uniquekeys.validate_master_league_query, uniquekeys.convert_region_to_platform)
    def get_league_summoner(self, query: Mapping[str, Any], context: PipelineContext = None) -> MasterLeague:
        return self._get(MasterLeague, query, uniquekeys.for_master_league_query, context)

    @get_many.register(MasterLeague)
    @validate_query(uniquekeys.validate_many_master_league_query, uniquekeys.convert_region_to_platform)
    def get_many_league_summoner(self, query: Mapping[str, Any], context: PipelineContext = None) -> Generator[MasterLeague, None, None]:
        return self._get_many(MasterLeague, query, uniquekeys.for_many_master_league_query, context)

    @put.register(MasterLeague)
    def put_league_summoner(self, item: MasterLeague, context: PipelineContext = None) -> None:
        self._put(MasterLeague, item, uniquekeys.for_master_league, context=context)

    @put_many.register(MasterLeague)
    def put_many_league_summoner(self, items: Iterable[MasterLeague], context: PipelineContext = None) -> None:
        self._put_many(MasterLeague, items, uniquekeys.for_master_league, context=context)

    ###################
    # Static Data API #
    ###################
    
    # Champion

    @get.register(Champion)
    @validate_query(uniquekeys.validate_champion_query, uniquekeys.convert_region_to_platform)
    def get_champion(self, query: Mapping[str, Any], context: PipelineContext = None) -> Champion:
        return self._get(Champion, query, uniquekeys.for_champion_query, context)

    @get_many.register(Champion)
    @validate_query(uniquekeys.validate_many_champion_query, uniquekeys.convert_region_to_platform)
    def get_many_champion(self, query: Mapping[str, Any], context: PipelineContext = None) -> Generator[Champion, None, None]:
        return self._get_many(Champion, query, uniquekeys.for_many_champion_query, context)

    @put.register(Champion)
    def put_champion(self, item: Champion, context: PipelineContext = None) -> None:
        self._put(Champion, item, uniquekeys.for_champion, ("id",), ("name",), context=context)

    @put_many.register(Champion)
    def put_many_champion(self, items: Iterable[Champion], context: PipelineContext = None) -> None:
        self._put_many(Champion, items, uniquekeys.for_champion, ("id",), ("name",), context=context)

    @get.register(Champions)
    @validate_query(uniquekeys.validate_champions_query, uniquekeys.convert_region_to_platform)
    def get_champions(self, query: Mapping[str, Any], context: PipelineContext = None) -> Champions:
        return self._get(Champions, query, uniquekeys.for_champions_query, context)

    @get_many.register(Champions)
    @validate_query(uniquekeys.validate_many_champions_query, uniquekeys.convert_region_to_platform)
    def get_many_champions(self, query: Mapping[str, Any], context: PipelineContext = None) -> Generator[Champions, None, None]:
        return self._get_many(Champions, query, uniquekeys.for_many_champions_query, context)

    @put.register(Champions)
    def put_champions(self, champions: Champions, context: PipelineContext = None) -> None:
        self._put(Champions, champions, uniquekeys.for_champions, ("platform",), context=context)
        for champion in champions:
            self._put(Champion, champion, uniquekeys.for_champion, ("id",), ("name",), context=context)

    @put_many.register(Champions)
    def put_many_champions(self, champions: Iterable[Champions], context: PipelineContext = None) -> None:
        self._put_many(Champions, champions, uniquekeys.for_champions, ("platform",), context=context)
        for champion in champions:
            self._put(Champion, champion, uniquekeys.for_champion, ("id",), ("name",), context=context)

    # Item

    @get.register(Item)
    @validate_query(uniquekeys.validate_item_query, uniquekeys.convert_region_to_platform)
    def get_item(self, query: Mapping[str, Any], context: PipelineContext = None) -> Item:
        return self._get(Item, query, uniquekeys.for_item_query, context)

    @get_many.register(Item)
    @validate_query(uniquekeys.validate_many_item_query, uniquekeys.convert_region_to_platform)
    def get_many_item(self, query: Mapping[str, Any], context: PipelineContext = None) -> Generator[Item, None, None]:
        return self._get_many(Item, query, uniquekeys.for_many_item_query, context)

    @put.register(Item)
    def put_item(self, item: Item, context: PipelineContext = None) -> None:
        self._put(Item, item, uniquekeys.for_item, ("id",), ("name",), context=context)

    @put_many.register(Item)
    def put_many_item(self, items: Iterable[Item], context: PipelineContext = None) -> None:
        self._put_many(Item, items, uniquekeys.for_item, ("id",), ("name",), context=context)

    @get.register(Items)
    @validate_query(uniquekeys.validate_items_query, uniquekeys.convert_region_to_platform)
    def get_items(self, query: Mapping[str, Any], context: PipelineContext = None) -> Items:
        return self._get(Items, query, uniquekeys.for_items_query, context)

    @get_many.register(Items)
    @validate_query(uniquekeys.validate_many_items_query, uniquekeys.convert_region_to_platform)
    def get_many_items(self, query: Mapping[str, Any], context: PipelineContext = None) -> Generator[Items, None, None]:
        return self._get_many(Items, query, uniquekeys.for_many_items_query, context)

    @put.register(Items)
    def put_items(self, items: Items, context: PipelineContext = None) -> None:
        self._put(Items, items, uniquekeys.for_items, ("platform",), context=context)
        for item in items:
            self._put(Item, item, uniquekeys.for_item, ("id",), ("name",), context=context)

    @put_many.register(Items)
    def put_many_items(self, items: Iterable[Items], context: PipelineContext = None) -> None:
        self._put_many(Items, items, uniquekeys.for_items, ("platform",), context=context)
        for item in items:
            self._put(Item, item, uniquekeys.for_item, ("id",), ("name",), context=context)

    # Language

    @get.register(Locales)
    @validate_query(uniquekeys.validate_languages_query, uniquekeys.convert_region_to_platform)
    def get_languages(self, query: Mapping[str, Any], context: PipelineContext = None) -> Locales:
        return self._get(Locales, query, uniquekeys.for_languages_query, context)

    @get_many.register(Locales)
    @validate_query(uniquekeys.validate_many_languages_query, uniquekeys.convert_region_to_platform)
    def get_many_languages(self, query: Mapping[str, Any], context: PipelineContext = None) -> Generator[Locales, None, None]:
        return self._get_many(Locales, query, uniquekeys.for_many_languages_query, context)

    @put.register(Locales)
    def put_languages(self, item: Locales, context: PipelineContext = None) -> None:
        self._put(Locales, item, uniquekeys.for_languages, context=context)

    @put_many.register(Locales)
    def put_many_languages(self, items: Iterable[Locales], context: PipelineContext = None) -> None:
        self._put_many(Locales, items, uniquekeys.for_languages, context=context)

    @get.register(LanguageStrings)
    @validate_query(uniquekeys.validate_language_strings_query, uniquekeys.convert_region_to_platform)
    def get_language_strings(self, query: Mapping[str, Any], context: PipelineContext = None) -> LanguageStrings:
        return self._get(LanguageStrings, query, uniquekeys.for_language_strings_query, context)

    @get_many.register(LanguageStrings)
    @validate_query(uniquekeys.validate_many_language_strings_query, uniquekeys.convert_region_to_platform)
    def get_many_language_strings(self, query: Mapping[str, Any], context: PipelineContext = None) -> Generator[LanguageStrings, None, None]:
        return self._get_many(LanguageStrings, query, uniquekeys.for_many_language_strings_query, context)

    @put.register(LanguageStrings)
    def put_language_strings(self, item: LanguageStrings, context: PipelineContext = None) -> None:
        self._put(LanguageStrings, item, uniquekeys.for_language_strings, context=context)

    @put_many.register(LanguageStrings)
    def put_many_language_strings(self, items: Iterable[LanguageStrings], context: PipelineContext = None) -> None:
        self._put_many(LanguageStrings, items, uniquekeys.for_language_strings, context=context)

    # Map

    @get.register(Map)
    @validate_query(uniquekeys.validate_map_query, uniquekeys.convert_region_to_platform)
    def get_map(self, query: Mapping[str, Any], context: PipelineContext = None) -> Map:
        return self._get(Map, query, uniquekeys.for_map_query, context)

    @get_many.register(Map)
    @validate_query(uniquekeys.validate_many_map_query, uniquekeys.convert_region_to_platform)
    def get_many_map(self, query: Mapping[str, Any], context: PipelineContext = None) -> Generator[Map, None, None]:
        return self._get_many(Map, query, uniquekeys.for_many_map_query, context)

    @put.register(Map)
    def put_map(self, item: Map, context: PipelineContext = None) -> None:
        self._put(Map, item, uniquekeys.for_map, ("id",), ("name",), context=context)

    @put_many.register(Map)
    def put_many_map(self, items: Iterable[Map], context: PipelineContext = None) -> None:
        self._put_many(Map, items, uniquekeys.for_map, ("id",), ("name",), context=context)

    @get.register(Maps)
    @validate_query(uniquekeys.validate_maps_query, uniquekeys.convert_region_to_platform)
    def get_maps(self, query: Mapping[str, Any], context: PipelineContext = None) -> Maps:
        return self._get(Maps, query, uniquekeys.for_maps_query, context)

    @get_many.register(Maps)
    @validate_query(uniquekeys.validate_many_maps_query, uniquekeys.convert_region_to_platform)
    def get_many_maps(self, query: Mapping[str, Any], context: PipelineContext = None) -> Generator[Maps, None, None]:
        return self._get_many(Maps, query, uniquekeys.for_many_maps_query, context)

    @put.register(Maps)
    def put_maps(self, item: Maps, context: PipelineContext = None) -> None:
        self._put(Maps, item, uniquekeys.for_maps, ("id",), ("name",), context=context)
        for map in item:
            self._put(Map, map, uniquekeys.for_map, ("id",), ("name",), context=context)

    @put_many.register(Maps)
    def put_many_maps(self, items: Iterable[Maps], context: PipelineContext = None) -> None:
        self._put_many(Maps, items, uniquekeys.for_maps, ("id",), ("name",), context=context)
        for maps in items:
            self._put_many(Map, maps, uniquekeys.for_map, ("id",), ("name",), context=context)

    # Mastery

    @get.register(Mastery)
    @validate_query(uniquekeys.validate_mastery_query, uniquekeys.convert_region_to_platform)
    def get_mastery(self, query: Mapping[str, Any], context: PipelineContext = None) -> Mastery:
        return self._get(Mastery, query, uniquekeys.for_mastery_query, context)

    @get_many.register(Mastery)
    @validate_query(uniquekeys.validate_many_mastery_query, uniquekeys.convert_region_to_platform)
    def get_many_mastery(self, query: Mapping[str, Any], context: PipelineContext = None) -> Generator[Mastery, None, None]:
        return self._get_many(Mastery, query, uniquekeys.for_many_mastery_query, context)

    @put.register(Mastery)
    def put_mastery(self, item: Mastery, context: PipelineContext = None) -> None:
        self._put(Mastery, item, uniquekeys.for_mastery, ("id",), ("name",), context=context)

    @put_many.register(Mastery)
    def put_many_mastery(self, items: Iterable[Mastery], context: PipelineContext = None) -> None:
        self._put_many(Mastery, items, uniquekeys.for_mastery, ("id",), ("name",), context=context)

    @get.register(Masteries)
    @validate_query(uniquekeys.validate_masteries_query, uniquekeys.convert_region_to_platform)
    def get_masteries(self, query: Mapping[str, Any], context: PipelineContext = None) -> Masteries:
        return self._get(Masteries, query, uniquekeys.for_masteries_query, context)

    @get_many.register(Masteries)
    @validate_query(uniquekeys.validate_many_masteries_query, uniquekeys.convert_region_to_platform)
    def get_many_masteries(self, query: Mapping[str, Any], context: PipelineContext = None) -> Generator[Masteries, None, None]:
        return self._get_many(Masteries, query, uniquekeys.for_many_masteries_query, context)

    @put.register(Masteries)
    def put_masteries(self, item: Masteries, context: PipelineContext = None) -> None:
        self._put(Masteries, item, uniquekeys.for_masteries, ("id",), ("name",), context=context)
        for mastery in item:
            self._put(Mastery, mastery, uniquekeys.for_mastery, ("id",), ("name",), context=context)

    @put_many.register(Masteries)
    def put_many_masteries(self, items: Iterable[Masteries], context: PipelineContext = None) -> None:
        self._put_many(Masteries, items, uniquekeys.for_masteries, ("id",), ("name",), context=context)
        for mastery in items:
            self._put(Mastery, mastery, uniquekeys.for_mastery, ("id",), ("name",), context=context)

    # Profile Icon

    @get.register(ProfileIcon)
    @validate_query(uniquekeys.validate_profile_icon_query, uniquekeys.convert_region_to_platform)
    def get_profile_icon(self, query: Mapping[str, Any], context: PipelineContext = None) -> ProfileIcon:
        return self._get(ProfileIcon, query, uniquekeys.for_profile_icon_query, context)

    @get_many.register(ProfileIcon)
    @validate_query(uniquekeys.validate_many_profile_icon_query, uniquekeys.convert_region_to_platform)
    def get_many_profile_icon(self, query: Mapping[str, Any], context: PipelineContext = None) -> Generator[ProfileIcon, None, None]:
        return self._get_many(ProfileIcon, query, uniquekeys.for_many_profile_icon_query, context)

    @put.register(ProfileIcon)
    def put_profile_icon(self, item: ProfileIcon, context: PipelineContext = None) -> None:
        self._put(ProfileIcon, item, uniquekeys.for_profile_icon, ("id",), ("name",), context=context)

    @put_many.register(ProfileIcon)
    def put_many_profile_icon(self, items: Iterable[ProfileIcon], context: PipelineContext = None) -> None:
        self._put_many(ProfileIcon, items, uniquekeys.for_profile_icon, ("id",), ("name",), context=context)

    @get.register(ProfileIcons)
    @validate_query(uniquekeys.validate_profile_icons_query, uniquekeys.convert_region_to_platform)
    def get_profile_icons(self, query: Mapping[str, Any], context: PipelineContext = None) -> ProfileIcons:
        return self._get(ProfileIcons, query, uniquekeys.for_profile_icons_query, context)

    @get_many.register(ProfileIcons)
    @validate_query(uniquekeys.validate_many_profile_icons_query, uniquekeys.convert_region_to_platform)
    def get_many_profile_icons(self, query: Mapping[str, Any], context: PipelineContext = None) -> Generator[ProfileIcons, None, None]:
        return self._get_many(ProfileIcons, query, uniquekeys.for_many_profile_icons_query, context)

    @put.register(ProfileIcons)
    def put_profile_icons(self, item: ProfileIcons, context: PipelineContext = None) -> None:
        self._put(ProfileIcons, item, uniquekeys.for_profile_icons, ("id",), ("name",), context=context)
        for profile_icon in item:
            self._put(ProfileIcon, profile_icon, uniquekeys.for_profile_icon, ("id",), ("name",), context=context)

    @put_many.register(ProfileIcons)
    def put_many_profile_icons(self, items: Iterable[ProfileIcons], context: PipelineContext = None) -> None:
        self._put_many(ProfileIcons, items, uniquekeys.for_profile_icons, ("id",), ("name",), context=context)
        for profile_icon in items:
            self._put(ProfileIcon, profile_icon, uniquekeys.for_profile_icon, ("id",), ("name",), context=context)

    # Realm

    @get.register(Realms)
    @validate_query(uniquekeys.validate_realms_query, uniquekeys.convert_region_to_platform)
    def get_realms(self, query: Mapping[str, Any], context: PipelineContext = None) -> Realms:
        return self._get(Realms, query, uniquekeys.for_realms_query, context)

    @get_many.register(Realms)
    @validate_query(uniquekeys.validate_many_realms_query, uniquekeys.convert_region_to_platform)
    def get_many_realms(self, query: Mapping[str, Any], context: PipelineContext = None) -> Generator[Realms, None, None]:
        return self._get_many(Realms, query, uniquekeys.for_many_realms_query, context)

    @put.register(Realms)
    def put_realms(self, item: Realms, context: PipelineContext = None) -> None:
        self._put(Realms, item, uniquekeys.for_realms, context=context)

    @put_many.register(Realms)
    def put_many_realms(self, items: Iterable[Realms], context: PipelineContext = None) -> None:
        self._put_many(Realms, items, uniquekeys.for_realms, context=context)

    # Rune

    @get.register(Rune)
    @validate_query(uniquekeys.validate_rune_query, uniquekeys.convert_region_to_platform)
    def get_rune(self, query: Mapping[str, Any], context: PipelineContext = None) -> Rune:
        return self._get(Rune, query, uniquekeys.for_rune_query, context)

    @get_many.register(Rune)
    @validate_query(uniquekeys.validate_many_rune_query, uniquekeys.convert_region_to_platform)
    def get_many_rune(self, query: Mapping[str, Any], context: PipelineContext = None) -> Generator[Rune, None, None]:
        return self._get_many(Rune, query, uniquekeys.for_many_rune_query, context)

    @put.register(Rune)
    def put_rune(self, item: Rune, context: PipelineContext = None) -> None:
        self._put(Rune, item, uniquekeys.for_rune, ("id",), ("name",), context=context)

    @put_many.register(Rune)
    def put_many_rune(self, items: Iterable[Rune], context: PipelineContext = None) -> None:
        self._put_many(Rune, items, uniquekeys.for_rune, ("id",), ("name",), context=context)

    @get.register(Runes)
    @validate_query(uniquekeys.validate_runes_query, uniquekeys.convert_region_to_platform)
    def get_runes(self, query: Mapping[str, Any], context: PipelineContext = None) -> Runes:
        return self._get(Runes, query, uniquekeys.for_runes_query, context)

    @get_many.register(Runes)
    @validate_query(uniquekeys.validate_many_runes_query, uniquekeys.convert_region_to_platform)
    def get_many_runes(self, query: Mapping[str, Any], context: PipelineContext = None) -> Generator[Runes, None, None]:
        return self._get_many(Runes, query, uniquekeys.for_many_runes_query, context)

    @put.register(Runes)
    def put_runes(self, item: Runes, context: PipelineContext = None) -> None:
        self._put(Runes, item, uniquekeys.for_runes, ("id",), ("name",), context=context)
        for rune in item:
            self._put(Mastery, rune, uniquekeys.for_rune, ("id",), ("name",), context=context)

    @put_many.register(Runes)
    def put_many_runes(self, items: Iterable[Runes], context: PipelineContext = None) -> None:
        self._put_many(Runes, items, uniquekeys.for_runes, ("id",), ("name",), context=context)
        for rune in items:
            self._put(Mastery, rune, uniquekeys.for_rune, ("id",), ("name",), context=context)

    # Summoner Spell

    @get.register(SummonerSpell)
    @validate_query(uniquekeys.validate_summoner_spell_query, uniquekeys.convert_region_to_platform)
    def get_summoner_spell(self, query: Mapping[str, Any], context: PipelineContext = None) -> SummonerSpell:
        return self._get(SummonerSpell, query, uniquekeys.for_summoner_spell_query, context)

    @get_many.register(SummonerSpell)
    @validate_query(uniquekeys.validate_many_summoner_spell_query, uniquekeys.convert_region_to_platform)
    def get_many_summoner_spell(self, query: Mapping[str, Any], context: PipelineContext = None) -> Generator[SummonerSpell, None, None]:
        return self._get_many(SummonerSpell, query, uniquekeys.for_many_summoner_spell_query, context)

    @put.register(SummonerSpell)
    def put_summoner_spell(self, item: SummonerSpell, context: PipelineContext = None) -> None:
        self._put(SummonerSpell, item, uniquekeys.for_summoner_spell, ("id",), ("name",), context=context)

    @put_many.register(SummonerSpell)
    def put_many_summoner_spell(self, items: Iterable[SummonerSpell], context: PipelineContext = None) -> None:
        self._put_many(SummonerSpell, items, uniquekeys.for_summoner_spell, ("id",), ("name",), context=context)

    @get.register(SummonerSpells)
    @validate_query(uniquekeys.validate_summoner_spells_query, uniquekeys.convert_region_to_platform)
    def get_summoner_spells(self, query: Mapping[str, Any], context: PipelineContext = None) -> SummonerSpells:
        return self._get(SummonerSpells, query, uniquekeys.for_summoner_spells_query, context)

    @get_many.register(SummonerSpells)
    @validate_query(uniquekeys.validate_many_summoner_spells_query, uniquekeys.convert_region_to_platform)
    def get_many_summoner_spells(self, query: Mapping[str, Any], context: PipelineContext = None) -> Generator[SummonerSpells, None, None]:
        return self._get_many(SummonerSpells, query, uniquekeys.for_many_summoner_spells_query, context)

    @put.register(SummonerSpells)
    def put_summoner_spells(self, item: SummonerSpells, context: PipelineContext = None) -> None:
        self._put(SummonerSpells, item, uniquekeys.for_summoner_spells, ("id",), ("name",), context=context)
        for summoner_spell in item:
            self._put(Mastery, summoner_spell, uniquekeys.for_summoner_spell, ("id",), ("name",), context=context)

    @put_many.register(SummonerSpells)
    def put_many_summoner_spells(self, items: Iterable[SummonerSpells], context: PipelineContext = None) -> None:
        self._put_many(SummonerSpells, items, uniquekeys.for_summoner_spells, ("id",), ("name",), context=context)
        for summoner_spell in items:
            self._put(Mastery, summoner_spell, uniquekeys.for_summoner_spell, ("id",), ("name",), context=context)

    # Versions

    @get.register(Versions)
    @validate_query(uniquekeys.validate_versions_query, uniquekeys.convert_region_to_platform)
    def get_versions(self, query: Mapping[str, Any], context: PipelineContext = None) -> Versions:
        return self._get(Versions, query, uniquekeys.for_versions_query, context=context)

    @get_many.register(Versions)
    @validate_query(uniquekeys.validate_many_versions_query, uniquekeys.convert_region_to_platform)
    def get_many_versions(self, query: Mapping[str, Any], context: PipelineContext = None) -> Generator[Versions, None, None]:
        return self._get_many(Versions, query, uniquekeys.for_many_versions_query, context=context)

    @put.register(Versions)
    def put_versions(self, item: Versions, context: PipelineContext = None) -> None:
        self._put(Versions, item, uniquekeys.for_versions, context=context)

    @put_many.register(Versions)
    def put_many_versions(self, items: Iterable[Versions], context: PipelineContext = None) -> None:
        self._put_many(Versions, items, uniquekeys.for_versions, context=context)

    ##############
    # Status API #
    ##############

    @get.register(ShardStatus)
    @validate_query(uniquekeys.validate_shard_status_query, uniquekeys.convert_region_to_platform)
    def get_shard_status(self, query: Mapping[str, Any], context: PipelineContext = None) -> ShardStatus:
        return self._get(ShardStatus, query, uniquekeys.for_shard_status_query, context)

    @get_many.register(ShardStatus)
    @validate_query(uniquekeys.validate_many_shard_status_query, uniquekeys.convert_region_to_platform)
    def get_many_shard_status(self, query: Mapping[str, Any], context: PipelineContext = None) -> Generator[ShardStatus, None, None]:
        return self._get_many(ShardStatus, query, uniquekeys.for_many_shard_status_query, context)

    @put.register(ShardStatus)
    def put_shard_status(self, item: ShardStatus, context: PipelineContext = None) -> None:
        self._put(ShardStatus, item, uniquekeys.for_shard_status, context=context)

    @put_many.register(ShardStatus)
    def put_many_shard_status(self, items: Iterable[ShardStatus], context: PipelineContext = None) -> None:
        self._put_many(ShardStatus, items, uniquekeys.for_shard_status, context=context)

    #################
    # Masteries API #
    #################

    @get.register(MasteryPage)
    @validate_query(uniquekeys.validate_mastery_page_query, uniquekeys.convert_region_to_platform)
    def get_mastery_page(self, query: Mapping[str, Any], context: PipelineContext = None) -> MasteryPage:
        return self._get(MasteryPage, query, uniquekeys.for_mastery_page_query, context)

    @get_many.register(MasteryPage)
    @validate_query(uniquekeys.validate_many_mastery_page_query, uniquekeys.convert_region_to_platform)
    def get_many_mastery_page(self, query: Mapping[str, Any], context: PipelineContext = None) -> Generator[MasteryPage, None, None]:
        return self._get_many(MasteryPage, query, uniquekeys.for_many_mastery_page_query, context)

    @put.register(MasteryPage)
    def put_mastery_page(self, item: MasteryPage, context: PipelineContext = None) -> None:
        self._put(MasteryPage, item, uniquekeys.for_mastery_page, ("id",), ("account.id",), ("name",), context=context)

    @put_many.register(MasteryPage)
    def put_many_mastery_page(self, items: Iterable[MasteryPage], context: PipelineContext = None) -> None:
        self._put_many(MasteryPage, items, uniquekeys.for_mastery_page, ("id",), ("account.id",), ("name",), context=context)

    #############
    # Match API #
    #############

    @get.register(Match)
    @validate_query(uniquekeys.validate_match_query, uniquekeys.convert_region_to_platform)
    def get_match(self, query: Mapping[str, Any], context: PipelineContext = None) -> Match:
        return self._get(Match, query, uniquekeys.for_match_query, context)

    @get_many.register(Match)
    @validate_query(uniquekeys.validate_many_match_query, uniquekeys.convert_region_to_platform)
    def get_many_match(self, query: Mapping[str, Any], context: PipelineContext = None) -> Generator[Match, None, None]:
        return self._get_many(Match, query, uniquekeys.for_many_match_query, context)

    @put.register(Match)
    def put_match(self, item: Match, context: PipelineContext = None) -> None:
        self._put(Match, item, uniquekeys.for_match, context=context)

    @put_many.register(Match)
    def put_many_match(self, items: Iterable[Match], context: PipelineContext = None) -> None:
        self._put_many(Match, items, uniquekeys.for_match, context=context)

    #############
    # Runes API #
    #############

    @get.register(RunePage)
    @validate_query(uniquekeys.validate_rune_page_query, uniquekeys.convert_region_to_platform)
    def get_rune_page(self, query: Mapping[str, Any], context: PipelineContext = None) -> RunePage:
        return self._get(RunePage, query, uniquekeys.for_rune_page_query, context)

    @get_many.register(RunePage)
    @validate_query(uniquekeys.validate_many_rune_page_query, uniquekeys.convert_region_to_platform)
    def get_many_rune_page(self, query: Mapping[str, Any], context: PipelineContext = None) -> Generator[RunePage, None, None]:
        return self._get_many(RunePage, query, uniquekeys.for_many_rune_page_query, context)

    @put.register(RunePage)
    def put_rune_page(self, item: RunePage, context: PipelineContext = None) -> None:
        self._put(RunePage, item, uniquekeys.for_rune_page, ("id",), ("account.id",), ("name",), context=context)

    @put_many.register(RunePage)
    def put_many_rune_page(self, items: Iterable[RunePage], context: PipelineContext = None) -> None:
        self._put_many(RunePage, items, uniquekeys.for_rune_page, ("id",), ("account.id",), ("name",), context=context)

    #################
    # Spectator API #
    #################

    @get.register(CurrentMatch)
    @validate_query(uniquekeys.validate_current_match_query, uniquekeys.convert_region_to_platform)
    def get_current_match(self, query: Mapping[str, Any], context: PipelineContext = None) -> CurrentMatch:
        return self._get(CurrentMatch, query, uniquekeys.for_current_match_query, context)

    @get_many.register(CurrentMatch)
    @validate_query(uniquekeys.validate_many_current_match_query, uniquekeys.convert_region_to_platform)
    def get_many_current_match(self, query: Mapping[str, Any], context: PipelineContext = None) -> Generator[CurrentMatch, None, None]:
        return self._get_many(CurrentMatch, query, uniquekeys.for_many_current_match_query, context)

    @put.register(CurrentMatch)
    def put_current_match(self, item: CurrentMatch, context: PipelineContext = None) -> None:
        self._put(CurrentMatch, item, uniquekeys.for_current_match, context=context)

    @put_many.register(CurrentMatch)
    def put_many_current_match(self, items: Iterable[CurrentMatch], context: PipelineContext = None) -> None:
        self._put_many(CurrentMatch, items, uniquekeys.for_current_match, context=context)

    @get.register(FeaturedMatches)
    @validate_query(uniquekeys.validate_featured_matches_query, uniquekeys.convert_region_to_platform)
    def get_featured_matches(self, query: Mapping[str, Any], context: PipelineContext = None) -> FeaturedMatches:
        return self._get(FeaturedMatches, query, uniquekeys.for_featured_matches_query, context)

    @get_many.register(FeaturedMatches)
    @validate_query(uniquekeys.validate_many_featured_matches_query, uniquekeys.convert_region_to_platform)
    def get_many_featured_matches(self, query: Mapping[str, Any], context: PipelineContext = None) -> Generator[FeaturedMatches, None, None]:
        return self._get_many(FeaturedMatches, query, uniquekeys.for_many_featured_matches_query, context)

    @put.register(FeaturedMatches)
    def put_featured_matches(self, item: FeaturedMatches, context: PipelineContext = None) -> None:
        self._put(FeaturedMatches, item, uniquekeys.for_featured_matches, context=context)

    @put_many.register(FeaturedMatches)
    def put_many_featured_matches(self, items: Iterable[FeaturedMatches], context: PipelineContext = None) -> None:
        self._put_many(FeaturedMatches, items, uniquekeys.for_featured_matches, context=context)

    ################
    # Summoner API #
    ################

    @get.register(Summoner)
    @validate_query(uniquekeys.validate_summoner_query, uniquekeys.convert_region_to_platform)
    def get_summoner(self, query: Mapping[str, Any], context: PipelineContext = None) -> Summoner:
        return self._get(Summoner, query, uniquekeys.for_summoner_query, context)

    @get_many.register(Summoner)
    @validate_query(uniquekeys.validate_many_summoner_query, uniquekeys.convert_region_to_platform)
    def get_many_summoner(self, query: Mapping[str, Any], context: PipelineContext = None) -> Generator[Summoner, None, None]:
        return self._get_many(Summoner, query, uniquekeys.for_many_summoner_query, context)

    @put.register(Summoner)
    def put_summoner(self, item: Summoner, context: PipelineContext = None) -> None:
        self._put(Summoner, item, uniquekeys.for_summoner, ("id",), ("account.id",), ("name",), context=context)

    @put_many.register(Summoner)
    def put_many_summoner(self, items: Iterable[Summoner], context: PipelineContext = None) -> None:
        self._put_many(Summoner, items, uniquekeys.for_summoner, ("id",), ("account.id",), ("name",), context=context)



import re
import zlib
from contextlib import contextmanager
from io import BytesIO
from typing import Mapping, MutableMapping, Any, Union, Dict
from urllib.parse import urlencode

from pycurl import Curl
from merakicommons.ratelimits import RateLimiter

try:
    import certifi
except ImportError:
    certifi = None

try:
    import ujson as json
except ImportError:
    import json


_print_calls = True


class HTTPError(RuntimeError):
    def __init__(self, message, code, response_headers: Dict[str, str] = None):
        super().__init__(message)
        self.code = code
        self.response_headers = response_headers or {}


class HTTPClient(object):
    @staticmethod
    def _execute(curl: Curl, close_connection: bool) -> int:
        curl.perform()
        status_code = curl.getinfo(curl.HTTP_CODE)
        if close_connection:
            curl.close()
        return status_code

    @staticmethod
    def _get(url: str, headers: Mapping[str, str] = None, rate_limiter: RateLimiter = None, connection: Curl = None) -> (int, bytes, dict):
        if not headers:
            request_headers = ["Accept-Encoding: gzip"]
        else:
            request_headers = ["{header}: {value}".format(header=key, value=value) for key, value in headers.items()]
            if "Accept-Encoding" not in headers:
                request_headers.append("Accept-Encoding: gzip")

        response_headers = {}

        def get_response_headers(header_line: bytes) -> None:
            header_line = header_line.decode("ISO-8859-1")

            if ":" not in header_line:
                return

            name, value = header_line.split(":", 1)
            response_headers[name.strip()] = value.strip()

        buffer = BytesIO()

        curl = connection if connection is not None else Curl()

        curl.setopt(curl.URL, url)
        curl.setopt(curl.WRITEDATA, buffer)
        curl.setopt(curl.HEADERFUNCTION, get_response_headers)
        curl.setopt(curl.HTTPHEADER, request_headers)
        if certifi:
            curl.setopt(curl.CAINFO, certifi.where())

        if _print_calls:  # TODO
            print("Making call: {}".format(url))
        if rate_limiter:
            with rate_limiter:
                status_code = HTTPClient._execute(curl, connection is None)
        else:
            status_code = HTTPClient._execute(curl, connection is None)

        body = buffer.getvalue()

        # Decompress if we got gzipped data
        try:
            content_encoding = response_headers["Content-Encoding"].upper()
            if "GZIP" == content_encoding:
                body = zlib.decompress(body, zlib.MAX_WBITS | 16)
        except KeyError:
            pass

        return status_code, body, response_headers

    def get(self, url: str, parameters: MutableMapping[str, Any] = None, headers: Mapping[str, str] = None, rate_limiter: RateLimiter = None, connection: Curl = None) -> (Union[dict, list, str, bytes], dict):
        if parameters:
            parameters = {k: str(v).lower() if isinstance(v, bool) else v for k, v in parameters.items()}
            parameters = urlencode(parameters, doseq=True)
            url = "{url}?{params}".format(url=url, params=parameters)

        status_code, body, response_headers = HTTPClient._get(url, headers, rate_limiter, connection)

        content_type = response_headers.get("Content-Type", "application/octet-stream").upper()

        # Decode to text if a charset is included
        match = re.search("CHARSET=(\S+)", content_type)
        if match:
            encoding = match.group(1)
            body = body.decode(encoding)

            # Load JSON if necessary
            if "APPLICATION/JSON" in content_type:
                body = json.loads(body)

        # Handle errors
        if status_code >= 400:
            if isinstance(body, dict):
                message = body.get("status", {}).get("message", "")
            elif isinstance(body, str):
                message = body
            else:
                message = ""

            raise HTTPError(message, status_code, response_headers)

        return body, response_headers

    @contextmanager
    def new_session(self) -> Curl:
        session = Curl()
        yield session
        session.close()



from typing import List, Union

from merakicommons.ghost import ghost_load_on
from merakicommons.cache import lazy
from merakicommons.container import searchable, SearchableList

from ..configuration import settings
from ..data import Region, Platform
from .common import CoreData, CassiopeiaObject, CassiopeiaGhost


##############
# Data Types #
##############


class TranslationData(CoreData):
    _renamed = {}

    @property
    def locale(self) -> str:
        return self._dto["locale"]

    @property
    def content(self) -> str:
        return self._dto["content"]

    @property
    def updated_at(self) -> str:
        return self._dto["updated_at"]


class MessageData(CoreData):
    _renamed = {}

    @property
    def severity(self) -> str:
        return self._dto["severity"]

    @property
    def author(self) -> str:
        return self._dto["author"]

    @property
    def created_at(self) -> str:
        return self._dto["created_at"]

    @property
    def translations(self) -> List[TranslationData]:
        return [TranslationData.from_dto(trans) for trans in self._dto["translations"]]

    @property
    def updated_at(self) -> str:
        return self._dto["updated_at"]

    @property
    def content(self) -> str:
        return self._dto["content"]

    @property
    def id(self) -> str:
        return self._dto["id"]


class IncidentData(CoreData):
    _renamed = {}

    @property
    def active(self) -> bool:
        return self._dto["active"]

    @property
    def created_at(self) -> str:
        return self._dto["created_at"]

    @property
    def id(self) -> int:
        return self._dto["id"]

    @property
    def updates(self) -> List[MessageData]:
        return [MessageData.from_dto(message) for message in self._dto["updates"]]


class ServiceData(CoreData):
    _renamed = {}

    @property
    def status(self) -> str:
        return self._dto["status"]

    @property
    def incidents(self) -> List[IncidentData]:
        return [IncidentData.from_dto(inc) for inc in self._dto["incidents"]]

    @property
    def name(self) -> str:
        return self._dto["name"]

    @property
    def slug(self) -> str:
        return self._dto["slug"]


class ShardStatusData(CoreData):
    _renamed = {}

    @property
    def region(self) -> str:
        return self._dto["region"]

    @property
    def name(self) -> str:
        return self._dto["name"]

    @property
    def region_tag(self) -> str:
        return self._dto["region_tag"]

    @property
    def hostname(self) -> str:
        return self._dto["hostname"]

    @property
    def services(self) -> List[ServiceData]:
        return [ServiceData.from_dto(service) for service in self._dto["services"]]

    @property
    def slug(self) -> str:
        return self._dto["slug"]

    @property
    def locales(self) -> List[str]:
        return self._dto["locales"]


##############
# Core Types #
##############


@searchable({})
class Translation(CassiopeiaObject):
    _data_types = {TranslationData}

    @property
    def locale(self) -> str:
        return self._data[TranslationData].locale

    @property
    def content(self) -> str:
        return self._data[TranslationData].content

    @property
    def updated_at(self) -> str:
        return self._data[TranslationData].updated_at


class Message(CassiopeiaObject):
    _data_types = {MessageData}

    @property
    def severity(self) -> str:
        return self._data[MessageData].severity

    @property
    def author(self) -> str:
        return self._data[MessageData].author

    @property
    def created_at(self) -> str:
        return self._data[MessageData].created_at

    @property
    def translations(self) -> List[Translation]:
        return SearchableList([Translation(trans) for trans in self._data[MessageData].translations])

    @property
    def updated_at(self) -> str:
        return self._data[MessageData].updated_at

    @property
    def content(self) -> str:
        return self._data[MessageData].content

    @property
    def id(self) -> str:
        return self._data[MessageData].id


class Incident(CassiopeiaObject):
    _data_types = {IncidentData}

    @property
    def active(self) -> bool:
        return self._data[IncidentData].active

    @property
    def created_at(self) -> str:
        return self._data[IncidentData].created_at

    @property
    def id(self) -> int:
        return self._data[IncidentData].id

    @property
    def updates(self) -> List[Message]:
        return SearchableList([Message.from_data(message) for message in self._data[IncidentData].updates])


class Service(CassiopeiaObject):
    _data_types = {ServiceData}

    @property
    def status(self) -> str:
        return self._data[ServiceData].status

    @property
    def incidents(self) -> List[Incident]:
        return SearchableList([Incident.from_data(inc) for inc in self._data[ServiceData].incidents])

    @property
    def name(self) -> str:
        return self._data[ServiceData].name

    @property
    def slug(self) -> str:
        return self._data[ServiceData].slug


@searchable({})
class ShardStatus(CassiopeiaGhost):
    _data_types = {ShardStatusData}

    def __init__(self, region: Union[Region, str] = None):
        if region is None:
            region = settings.default_region
        if not isinstance(region, Region):
            region = Region(region)
        kwargs = {"region": region}
        super().__init__(**kwargs)

    def __get_query__(self):
        return {"region": self.region, "platform": self.platform}

    @CassiopeiaGhost.property(ShardStatusData)
    @ghost_load_on(KeyError)
    @lazy
    def region(self) -> Region:
        return Region(self._data[ShardStatusData].region)

    @CassiopeiaGhost.property(ShardStatusData)
    @ghost_load_on(KeyError)
    @lazy
    def platform(self) -> Platform:
        return self.region.platform

    @CassiopeiaGhost.property(ShardStatusData)
    @ghost_load_on(KeyError)
    def name(self) -> str:
        return self._data[ShardStatusData].name

    @CassiopeiaGhost.property(ShardStatusData)
    @ghost_load_on(KeyError)
    def region_tag(self) -> str:
        return self._data[ShardStatusData].region_tag

    @CassiopeiaGhost.property(ShardStatusData)
    @ghost_load_on(KeyError)
    def hostname(self) -> str:
        return self._data[ShardStatusData].hostname

    @CassiopeiaGhost.property(ShardStatusData)
    @ghost_load_on(KeyError)
    def services(self) -> List[Service]:
        return SearchableList([Service.from_data(service) for service in self._data[ShardStatusData].services])

    @CassiopeiaGhost.property(ShardStatusData)
    @ghost_load_on(KeyError)
    def slug(self) -> str:
        return self._data[ShardStatusData].slug

    @CassiopeiaGhost.property(ShardStatusData)
    @ghost_load_on(KeyError)
    def locales(self) -> List[str]:
        return self._data[ShardStatusData].locales



import datetime
from typing import Union

from merakicommons.ghost import ghost_load_on
from merakicommons.cache import lazy, lazy_property
from merakicommons.container import searchable, SearchableList

from ..configuration import settings
from ..data import Region, Platform
from .common import CoreData, CassiopeiaGhost, CassiopeiaGhostList, DataObjectList
from ..dto.championmastery import ChampionMasteryDto
from .staticdata.champion import Champion
from .summoner import Summoner
from ..dto import championmastery as dto


##############
# Data Types #
##############


class ChampionMasteryListData(DataObjectList):
    _dto_type = dto.ChampionMasteryListDto
    _renamed = {"summoner_id": "summonerId"}

    @property
    def region(self) -> str:
        return self._dto["region"]

    @property
    def summoner_id(self) -> str:
        return self._dto["summonerId"]


class ChampionMasteryData(CoreData):
    _dto_type = ChampionMasteryDto
    _renamed = {"chest_granted": "chestGranted", "level": "championLevel", "points": "championPoints", "champion_id": "championId", "summoner_id": "playerId", "points_until_next_level": "championPointsUntilNextLevel", "points_since_last_level": "championPointsSinceLastLevel", "last_played": "lastPlayTime"}

    @property
    def region(self) -> str:
        return self._dto["region"]

    @property
    def champion_id(self) -> int:
        return self._dto["championId"]

    @property
    def summoner_id(self) -> int:
        return self._dto["playerId"]

    @property
    def chest_granted(self) -> bool:
        return self._dto["chestGranted"]

    @property
    def level(self) -> int:
        return self._dto["championLevel"]

    @property
    def points(self) -> int:
        return self._dto["championPoints"]

    @property
    def points_until_next_level(self) -> int:
        return self._dto["championPointsUntilNextLevel"]

    @property
    def points_since_last_level(self) -> int:
        return self._dto["championPointsSinceLastLevel"]

    @property
    def last_played(self) -> int:
        return self._dto["lastPlayTime"]


##############
# Core Types #
##############


class ChampionMasteries(CassiopeiaGhostList):
    _data_types = {ChampionMasteryListData}

    def __init__(self, *args, summoner: Union[Summoner, int, str], region: Union[Region, str] = None):
        super().__init__(*args, region=region)
        if isinstance(summoner, str):
            summoner = Summoner(name=summoner)
        elif isinstance(summoner, int):
            summoner = Summoner(id=summoner)
        self.__summoner = summoner

    def __get_query__(self) -> dict:
        return {"region": self.region, "summoner.id": self.summoner.id}

    def __load_hook__(self, load_group: CoreData, data: CoreData) -> None:
        self.clear()
        from ..transformers.championmastery import ChampionMasteryTransformer
        SearchableList.__init__(self, [ChampionMasteryTransformer.champion_mastery_data_to_core(None, cm) for cm in data])
        super().__load_hook__(load_group, data)

    @lazy_property
    def region(self) -> Region:
        return Region(self._data[ChampionMasteryListData].region)

    @lazy_property
    def platform(self) -> Platform:
        return self.region.platform

    @property
    def summoner(self):
        return self.__summoner


@searchable({str: ["champion", "summoner"], int: ["points", "level"], bool: ["chest_granted"], datetime.datetime: ["last_played"], Champion: ["champion"], Summoner: ["summoner"]})
class ChampionMastery(CassiopeiaGhost):
    _data_types = {ChampionMasteryData}

    def __init__(self, *, summoner: Union[Summoner, int, str] = None, champion: Union[Champion, int, str] = None, region: Union[Region, str] = None):
        if region is None:
            region = settings.default_region
        if not isinstance(region, Region):
            region = Region(region)
        kwargs = {"region": region}

        if summoner is not None:
            if isinstance(summoner, Summoner):
                self.__class__.summoner.fget._lazy_set(self, summoner)
            elif isinstance(summoner, str):
                summoner = Summoner(name=summoner)
                self.__class__.summoner.fget._lazy_set(self, summoner)
            else:  # int
                kwargs["summoner_id"] = summoner

        if champion is not None:
            if isinstance(champion, Champion):
                self.__class__.champion.fget._lazy_set(self, champion)
            elif isinstance(champion, str):
                champion = Champion(name=champion)
                self.__class__.champion.fget._lazy_set(self, champion)
            else:  # int
                kwargs["champion_id"] = champion

        super().__init__(**kwargs)

    def __get_query__(self):
        return {"region": self.region, "platform": self.platform.value, "summoner.id": self.summoner.id, "champion.id": self.champion.id}

    def __load__(self, load_group: CoreData = None) -> None:
        from datapipelines import NotFoundError
        try:
            return super().__load__(load_group)
        except NotFoundError:
            from ..transformers.championmastery import ChampionMasteryTransformer
            dto = {
                "championLevel": 0,
                "chestGranted": False,
                "championPoints": 0,
                "championPointsUntilNextLevel": 1800,
                "championPointsSinceLastLevel": 0,
                "lastPlayTime": None
            }
            data = ChampionMasteryTransformer.champion_mastery_dto_to_data(None, dto)
            self.__load_hook__(load_group, data)

    @property
    def region(self) -> Region:
        return Region(self._data[ChampionMasteryData].region)

    @lazy_property
    def platform(self) -> Platform:
        return self.region.platform

    @CassiopeiaGhost.property(ChampionMasteryData)
    @lazy
    def champion(self) -> Champion:
        """Champion for this entry."""
        return Champion(id=self._data[ChampionMasteryData].champion_id)

    @CassiopeiaGhost.property(ChampionMasteryData)
    @lazy
    def summoner(self) -> Summoner:
        """Summoner for this entry."""
        return Summoner(id=self._data[ChampionMasteryData].summoner_id)

    @CassiopeiaGhost.property(ChampionMasteryData)
    @ghost_load_on(KeyError)
    def chest_granted(self) -> bool:
        """Is chest granted for this champion or not in current season?"""
        return self._data[ChampionMasteryData].chest_granted

    @CassiopeiaGhost.property(ChampionMasteryData)
    @ghost_load_on(KeyError)
    def level(self) -> int:
        """Champion level for specified player and champion combination."""
        return self._data[ChampionMasteryData].level

    @CassiopeiaGhost.property(ChampionMasteryData)
    @ghost_load_on(KeyError)
    def points(self) -> int:
        """Total number of champion points for this player and champion combination - used to determine champion level."""
        return self._data[ChampionMasteryData].points

    @CassiopeiaGhost.property(ChampionMasteryData)
    @ghost_load_on(KeyError)
    def points_until_next_level(self) -> int:
        """Number of points needed to achieve next level. Zero if player reached maximum champion level for this champion."""
        return self._data[ChampionMasteryData].points_until_next_level

    @CassiopeiaGhost.property(ChampionMasteryData)
    @ghost_load_on(KeyError)
    def points_since_last_level(self) -> int:
        """Number of points earned since current level has been achieved. Zero if player reached maximum champion level for this champion."""
        return self._data[ChampionMasteryData].points_since_last_level

    @CassiopeiaGhost.property(ChampionMasteryData)
    @ghost_load_on(KeyError)
    @lazy
    def last_played(self) -> datetime.datetime:
        """Last time this champion was played by this player."""
        return datetime.datetime.fromtimestamp(self._data[ChampionMasteryData].last_played / 1000)



from enum import Enum
from collections import OrderedDict

from .patches import patches as _patches

PATCHES = OrderedDict(((patch.start, patch) for patch in _patches), key=lambda patch: patch.start)


class Region(Enum):
    brazil = "BR"
    europe_north_east = "EUNE"
    europe_west = "EUW"
    japan = "JP"
    korea = "KR"
    latin_america_north = "LAN"
    latin_america_south = "LAS"
    north_america = "NA"
    oceania = "OCE"
    turkey = "TR"
    russia = "RU"
    pbe = "PBE"

    @property
    def platform(self) -> "Platform":
        return getattr(Platform, self.name)

    @property
    def default_locale(self) -> str:
        return DEFAULT_LOCALE[self]


class Platform(Enum):
    brazil = "BR1"
    europe_north_east = "EUN1"
    europe_west = "EUW1"
    japan = "JP1"
    korea = "KR"
    latin_america_north = "LA1"
    latin_america_south = "LA2"
    north_america = "NA1"
    oceania = "OC1"
    turkey = "TR1"
    russia = "RU"
    pbe = "PBE1"

    @property
    def region(self) -> "Region":
        return getattr(Region, self.name)

    @property
    def default_locale(self) -> str:
        return DEFAULT_LOCALE[self]


DEFAULT_LOCALE = {
    Region.brazil: "pt_BR",
    Platform.brazil: "pt_BR",
    Region.europe_north_east: "en_GB",
    Platform.europe_north_east: "en_GB",
    Region.europe_west: "en_GB",
    Platform.europe_west: "en_GB",
    Region.japan: "ja_JP",
    Platform.japan: "ja_JP",
    Region.korea: "ko_KR",
    Platform.korea: "ko_KR",
    Region.latin_america_north: "es_MX",
    Platform.latin_america_north: "es_MX",
    Region.latin_america_south: "es_AR",
    Platform.latin_america_south: "es_AR",
    Region.north_america: "en_US",
    Platform.north_america: "en_US",
    Region.oceania: "en_AU",
    Platform.oceania: "en_AU",
    Region.turkey: "tr_TR",
    Platform.turkey: "tr_TR",
    Region.russia: "ru_RU",
    Platform.russia: "ru_RU",
    Region.pbe: "en_US",
    Platform.pbe: "en_US"
}


class Resource(Enum):
    mana = "Mana"
    courage = "Courage"
    energy = "Energy"
    fury = "Fury"
    rage = "Rage"
    flow = "Flow"
    ferocity = "Ferocity"
    heat = "Heat"
    shield = "Shield"
    blood_well = "Blood Well"
    crimson_rush = "Crimson Rush"
    none = "None"
    no_cost = "No Cost"


class Side(Enum):
    blue = 100
    red = 200


class Map(Enum):
    summoners_rift_summer = 1
    summoners_rift_autumn = 2
    the_proving_grounds = 3
    twisted_treeline_original = 4
    the_crystal_scar = 8
    twisted_treeline = 10
    summoners_rift = 11
    howling_abyss = 12
    butchers_bridge = 14
    cosmic_ruins = 16


class GameMode(Enum):
    aram = "ARAM"
    ascension = "ASCENSION"
    classic = "CLASSIC"
    showdown = "FIRSTBLOOD"
    poro_king = "KINGPORO"
    dominion = "ODIN"
    one_for_all = "ONEFORALL"
    tutorial = "TUTORIAL"
    nexus_siege = "SIEGE"
    assassinate = "ASSASSINATE"
    dark_star = "DARKSTAR"
    arsr = "ARSR"
    urf = "URF"
    doom_bots = "DOOMBOTSTEEMO"


class MasteryTree(Enum):
    cunning = "Cunning"
    ferocity = "Ferocity"
    resolve = "Resolve"


class RuneType(Enum):
    mark = "red"
    seal = "yellow"
    glyph = "blue"
    quint = "black"


class Tier(Enum):
    challenger = "CHALLENGER"
    master = "MASTER"
    diamond = "DIAMOND"
    platinum = "PLATINUM"
    gold = "GOLD"
    silver = "SILVER"
    bronze = "BRONZE"
    unranked = "UNRANKED"


class Division(Enum):
    one = "I"
    two = "II"
    three = "III"
    four = "IV"
    five = "V"


class Season(Enum):
    preseason_3 = "PRESEASON3"
    season_3 = "SEASON3"
    preseason_4 = "PRESEASON2014"
    season_4 = "SEASON2014"
    preseason_5 = "PRESEASON2015"
    season_5 = "SEASON2015"
    preseason_6 = "PRESEASON2016"
    season_6 = "SEASON2016"
    preseason_7 = "PRESEASON2017"
    season_7 = "SEASON2017"

    def from_id(id: int):
        return {i: season for season, i in SEASON_IDS.items()}[id]


SEASON_IDS = {
    Season.preseason_3: 0,
    Season.season_3: 1,
    Season.preseason_4: 2,
    Season.season_4: 3,
    Season.preseason_5: 4,
    Season.season_5: 5,
    Season.preseason_6: 6,
    Season.season_6: 7,
    Season.preseason_7: 8,
    Season.season_7: 9
}


class GameType(Enum):
    custom = "CUSTOM_GAME"
    tutorial = "TUTORIAL_GAME"
    matched = "MATCHED_GAME"


class Lane(Enum):
    top_lane = "TOP_LANE"
    mid_lane = "MID_LANE"
    bot_lane = "BOT_LANE"


class Queue(Enum):
    custom = "CUSTOM"
    normal_blind_threes = "NORMAL_3x3"
    normal_blind_fives = "NORMAL_5x5_BLIND"
    normal_draft_fives = "NORMAL_5x5_DRAFT"
    ranked_solo = "RANKED_SOLO_5x5"
    ranked_premade_fives = "RANKED_PREMADE_5x5"
    ranked_premade_threes = "RANKED_PREMADE_3x3"
    flex_threes = "RANKED_FLEX_TT"
    ranked_threes = "RANKED_TEAM_3x3"
    ranked_fives = "RANKED_TEAM_5x5"
    dominion_blind = "ODIN_5x5_BLIND"
    dominion_draft = "ODIN_5x5_DRAFT"
    bot_fives = "BOT_5x5"
    bot_dominion = "BOT_ODIN_5x5"
    bot_intro_fives = "BOT_5x5_INTRO"
    bot_beginner_fives = "BOT_5x5_BEGINNER"
    bot_intermediate_fives = "BOT_5x5_INTERMEDIATE"
    bot_threes = "BOT_TT_3x3"
    team_builder = "GROUP_FINDER_5x5"
    aram = "ARAM_5x5"
    one_for_all = "ONEFORALL_5x5"
    showdown_solo = "FIRSTBLOOD_1x1"
    showdown_duo = "FIRSTBLOOD_2x2"
    hexakill_summoners_rift = "SR_6x6"
    urf = "URF_5x5"
    one_for_all_mirror = "ONEFORALL_MIRRORMODE_5x5"
    bot_urf = "BOT_URF_5x5"
    doom_bots_1 = "NIGHTMARE_BOT_5x5_RANK1"
    doom_bots_2 = "NIGHTMARE_BOT_5x5_RANK2"
    doom_bots_5 = "NIGHTMARE_BOT_5x5_RANK5"
    ascension = "ASCENSION_5x5"
    hexakill_twisted_treeline = "HEXAKILL"
    butchers_bridge = "BILGEWATER_ARAM_5x5"
    poro_king = "KING_PORO_5x5"
    nemesis_draft = "COUNTER_PICK"
    black_market = "BILGEWATER_5x5"
    nexus_siege = "SIEGE"
    definitely_not_dominion = "DEFINITELY_NOT_DOMINION_5x5"
    all_random_urf = "ARURF_5X5"
    all_random_summoners_rift = "ARSR_5x5"
    dynamic_queue = "TEAM_BUILDER_DRAFT_UNRANKED_5x5"
    ranked_dynamic_queue = "TEAM_BUILDER_DRAFT_RANKED_5x5"
    ranked_solo_queue = "TEAM_BUILDER_RANKED_SOLO"
    flex = "RANKED_FLEX_SR"
    blood_hunt = "ASSASSINATE_5x5"
    darkstar = "DARKSTAR_3x3"

    def from_id(id: int):
        return {i: season for season, i in QUEUE_IDS.items()}[id]


QUEUE_IDS = {
    Queue.custom: 0,
    Queue.normal_blind_threes: 8,
    Queue.normal_blind_fives: 2,
    Queue.normal_draft_fives: 14,
    Queue.ranked_solo: 4,
    Queue.ranked_premade_fives: 6,
    Queue.ranked_premade_threes: 9,
    Queue.flex_threes: 9,
    Queue.ranked_threes: 41,
    Queue.ranked_fives: 42,
    Queue.dominion_blind: 16,
    Queue.dominion_draft: 17,
    Queue.bot_fives: 7,
    Queue.bot_dominion: 25,
    Queue.bot_intro_fives: 31,
    Queue.bot_beginner_fives: 32,
    Queue.bot_intermediate_fives: 33,
    Queue.bot_threes: 52,
    Queue.team_builder: 61,
    Queue.aram: 65,
    Queue.one_for_all: 70,
    Queue.showdown_solo: 72,
    Queue.showdown_duo: 73,
    Queue.hexakill_summoners_rift: 75,
    Queue.urf: 76,
    Queue.one_for_all_mirror: 78,
    Queue.bot_urf: 83,
    Queue.doom_bots_1: 91,
    Queue.doom_bots_2: 92,
    Queue.doom_bots_5: 93,
    Queue.ascension: 96,
    Queue.hexakill_twisted_treeline: 98,
    Queue.butchers_bridge: 100,
    Queue.poro_king: 300,
    Queue.nemesis_draft: 310,
    Queue.black_market: 313,
    Queue.nexus_siege: 315,
    Queue.definitely_not_dominion: 317,
    Queue.all_random_urf: 318,
    Queue.all_random_summoners_rift: 325,
    Queue.dynamic_queue: 400,
    Queue.ranked_dynamic_queue: 410,
    Queue.ranked_solo_queue: 420,
    Queue.flex: 440,
    Queue.blood_hunt: 600,
    Queue.darkstar: 610
}

RANKED_QUEUES = {
    Queue.flex,
    Queue.flex_threes,
    Queue.ranked_solo
}



from typing import Set, Union

from merakicommons.ghost import ghost_load_on
from merakicommons.cache import lazy, lazy_property
from merakicommons.container import searchable, SearchableList

from ...configuration import settings
from ...data import Region, Platform, MasteryTree
from ..common import CoreData, DataObjectList, CassiopeiaGhost, CassiopeiaGhostList, get_latest_version
from .common import Sprite, Image
from ...dto.staticdata import mastery as dto


##############
# Data Types #
##############


class MasteryListData(DataObjectList):
    _dto_type = dto.MasteryListDto
    _renamed = {"included_data": "includedData"}

    @property
    def region(self) -> str:
        return self._dto["region"]

    @property
    def version(self) -> str:
        return self._dto["version"]

    @property
    def locale(self) -> str:
        return self._dto["locale"]

    @property
    def included_data(self) -> Set[str]:
        return self._dto["includedData"]


class MasteryData(CoreData):
    _dto_type = dto.MasteryDto
    _renamed = {"prerequisite": "prereq", "tree": "masteryTree", "image_data": "image", "points": "ranks", "sanitized_description": "sanitizedDescription", "included_data": "includedData"}

    @property
    def region(self) -> str:
        return self._dto["region"]

    @property
    def version(self) -> str:
        return self._dto["version"]

    @property
    def locale(self) -> str:
        return self._dto["locale"]

    @property
    def included_data(self) -> Set[str]:
        return self._dto["includedData"]

    @property
    def prerequisite(self) -> id:
        """Always '0'"""
        return self._dto["prereq"]

    @property
    def tree(self) -> str:
        """Legal values: Cunning, Ferocity, Resolve"""
        return self._dto["masteryTree"]

    @property
    def name(self) -> str:
        return self._dto["name"]

    @property
    def points(self) -> list:
        return self._dto["ranks"]

    @property
    def image_data(self) -> str:
        return self._dto["image"]

    @property
    def id(self) -> str:
        return self._dto["id"]

    @property
    def sanitized_description(self) -> str:
        return self._dto["sanitizedDescription"]

    @property
    def description(self) -> str:
        return self._dto["description"]


##############
# Core Types #
##############


class Masteries(CassiopeiaGhostList):
    _data_types = {MasteryListData}

    def __init__(self, *args, region: Union[Region, str] = None, version: str = None, locale: str = None, included_data: Set[str] = None):
        if region is None:
            region = settings.default_region
        if not isinstance(region, Region):
            region = Region(region)
        if included_data is None:
            included_data = {"all"}
        if locale is None:
            locale = region.default_locale
        kwargs = {"region": region, "included_data": included_data, "locale": locale}
        if version:
            kwargs["version"] = version
        super().__init__(*args, **kwargs)

    def __get_query__(self):
        return {"region": self.region, "platform": self.platform, "version": self.version, "locale": self.locale, "includedData": self.included_data}

    def __load_hook__(self, load_group: CoreData, data: CoreData) -> None:
        self.clear()
        from ...transformers.staticdata import StaticDataTransformer
        SearchableList.__init__(self, [StaticDataTransformer.mastery_data_to_core(None, i) for i in data])
        super().__load_hook__(load_group, data)

    @lazy_property
    def region(self) -> Region:
        return Region(self._data[MasteryListData].region)

    @lazy_property
    def platform(self) -> Platform:
        return self.region.platform

    @property
    def version(self) -> str:
        try:
            return self._data[MasteryListData].version
        except KeyError:
            version = get_latest_version(region=self.region, endpoint="mastery")
            self(version=version)
            return self._data[MasteryListData].version

    @property
    def locale(self) -> str:
        return self._data[MasteryListData].locale

    @property
    def included_data(self) -> Set[str]:
        """A set of tags to return additional information for this mastery when it's loaded."""
        return self._data[MasteryListData].included_data


@searchable({str: ["name", "key", "region", "platform", "locale", "tree"], int: ["id"], MasteryTree: ["tree"], Region: ["region"], Platform: ["platform"]})
class Mastery(CassiopeiaGhost):
    _data_types = {MasteryData}
    _load_types = {MasteryData: MasteryListData}

    def __init__(self, *, id: int = None, name: str = None, region: Union[Region, str] = None, version: str = None, locale: str = None, included_data: Set[str] = None):
        if region is None:
            region = settings.default_region
        if not isinstance(region, Region):
            region = Region(region)
        if included_data is None:
            included_data = {"all"}
        if locale is None:
            locale = region.default_locale
        kwargs = {"region": region, "included_data": included_data, "locale": locale}
        if id is not None:
            kwargs["id"] = id
        if name is not None:
            kwargs["name"] = name
        if version is not None:
            kwargs["version"] = version
        super().__init__(**kwargs)

    def __get_query__(self):
        return {"region": self.region, "platform": self.platform, "version": self.version, "locale": self.locale, "includedData": self.included_data}

    def __load_hook__(self, load_group: CoreData, data: CoreData) -> None:
        def find_matching_attribute(datalist, attrname, attrvalue):
            for item in datalist:
                if getattr(item, attrname, None) == attrvalue:
                    return item

        # The `data` is a dict of summoner spell data instances
        if "id" in self._data[MasteryData]._dto:
            find = "id", self.id
        elif "name" in self._data[MasteryData]._dto:
            find = "name", self.name
        else:
            raise RuntimeError("Expected fields not present after loading.")
        data = find_matching_attribute(data, *find)

        super().__load_hook__(load_group, data)

    @lazy_property
    def region(self) -> Region:
        """The region for this mastery."""
        return Region(self._data[MasteryData].region)

    @lazy_property
    def platform(self) -> Platform:
        """The platform for this mastery."""
        return self.region.platform

    @property
    def version(self) -> str:
        """The version for this mastery."""
        try:
            return self._data[MasteryData].version
        except KeyError:
            version = get_latest_version(region=self.region, endpoint="mastery")
            self(version=version)
            return self._data[MasteryData].version

    @property
    def locale(self) -> str:
        """The locale for this mastery."""
        return self._data[MasteryData].locale

    @property
    def included_data(self) -> Set[str]:
        """A set of tags to return additional information for this mastery when it's loaded."""
        return self._data[MasteryData].included_data

    @CassiopeiaGhost.property(MasteryData)
    @ghost_load_on(KeyError)
    def prerequisite(self) -> "Mastery":
        """The prerequisite masteries."""
        return Mastery(id=self._data[MasteryData].prerequisite)

    @CassiopeiaGhost.property(MasteryData)
    @ghost_load_on(KeyError)
    def tree(self) -> MasteryTree:
        """The mastery's tree (Cunning, Ferocity, Resolve)."""
        return MasteryTree(self._data[MasteryData].tree)

    @CassiopeiaGhost.property(MasteryData)
    @ghost_load_on(KeyError)
    def name(self) -> str:
        """The mastery's name."""
        return self._data[MasteryData].name

    @CassiopeiaGhost.property(MasteryData)
    @ghost_load_on(KeyError)
    def points(self) -> str:
        """The number of points that can be put into this mastery."""
        return self._data[MasteryData].points

    @CassiopeiaGhost.property(MasteryData)
    @ghost_load_on(KeyError)
    def id(self) -> int:
        """The mastery's ID."""
        return self._data[MasteryData].id

    @CassiopeiaGhost.property(MasteryData)
    @ghost_load_on(KeyError)
    def description(self) -> int:
        """The mastery's description."""
        return self._data[MasteryData].description

    @CassiopeiaGhost.property(MasteryData)
    @ghost_load_on(KeyError)
    def sanitized_description(self) -> int:
        """The mastery's sanitized description."""
        return self._data[MasteryData].sanitized_description

    @CassiopeiaGhost.property(MasteryData)
    @ghost_load_on(KeyError)
    def id(self) -> int:
        """The mastery's ID."""
        return self._data[MasteryData].id

    @CassiopeiaGhost.property(MasteryData)
    @ghost_load_on(KeyError)
    @lazy
    def image(self) -> Image:
        """The image information for this mastery."""
        image = Image.from_data(self._data[MasteryData].image)
        image(version=self.version)
        return image

    @lazy_property
    def sprite(self) -> Sprite:
        return self.image.sprite_info



from typing import Dict, List, Set, Union

from PIL.Image import Image as PILImage
from merakicommons.ghost import ghost_load_on
from merakicommons.cache import lazy, lazy_property
from merakicommons.container import searchable, SearchableList, SearchableDictionary

from ...configuration import settings
from ...data import Resource, Region, Platform, Map, GameMode
from ..champion import ChampionData as ChampionStatusData
from ..champion import ChampionListData as ChampionStatusListData
from ..common import CoreData, CassiopeiaObject, CassiopeiaGhost, CassiopeiaGhostList, DataObjectList, get_latest_version
from .common import ImageData, Image, Sprite
from ...dto.staticdata import champion as dto
from .item import Item


##############
# Data Types #
##############


class ChampionListData(DataObjectList):
    _dto_type = dto.ChampionListDto
    _renamed = {"included_data": "includedData"}

    @property
    def region(self) -> str:
        return self._dto["region"]

    @property
    def version(self) -> str:
        return self._dto["version"]

    @property
    def locale(self) -> str:
        return self._dto["locale"]

    @property
    def included_data(self) -> Set[str]:
        """A set of tags to return additonal information for this champion when it's loaded."""
        return self._dto["includedData"]


class SpellVarsData(CoreData):
    _renamed = {"ranks_with": "ranksWith", "dynamic": "dyn", "coefficients": "coeff"}

    @property  # This doesn't get returned by the API
    def ranks_with(self) -> str:
        return self._dto["ranksWith"]

    @property  # This doesn't get returned by the API
    def dynamic(self) -> str:
        return self._dto["dyn"]

    @property
    def link(self) -> str:
        return self._dto["link"]

    @property
    def coefficients(self) -> List[float]:
        return self._dto["coeff"]

    @property
    def key(self) -> str:
        return self._dto["key"]


class LevelTipData(CoreData):
    _renamed = {"effects": "effect", "keywords": "label"}

    @property
    def effects(self) -> List[str]:
        return self._dto["effect"]

    @property
    def keywords(self) -> List[str]:
        return self._dto["label"]


class ChampionSpellData(CoreData):
    _renamed = {"level_up_tips": "leveltip", "variables": "vars", "sanitized_description": "sanitizedDescription", "sanitized_tooltip": "sanitizedTooltip",
                "max_rank": "maxrank", "cooldowns": "cooldown", "costs": "cost", "alternative_images": "altimages", "effects": "effect", "resource": "costType"}

    @property
    def level_up_tips(self) -> LevelTipData:
        return LevelTipData.from_dto(self._dto["leveltip"])

    @property
    def variables(self) -> List[SpellVarsData]:
        return [SpellVarsData.from_dto(v) for v in self._dto["vars"]]

    @property
    def resource(self) -> str:
        return self._dto["costType"]

    @property
    def image(self) -> ImageData:
        return ImageData.from_dto(self._dto["image"])

    @property
    def sanitized_description(self) -> str:
        return self._dto["sanitizedDescription"]

    @property
    def sanitized_tooltip(self) -> str:
        return self._dto["sanitizedTooltip"]

    @property
    def effects(self) -> List[List[float]]:
        return self._dto["effect"]

    @property
    def tooltip(self) -> str:
        return self._dto["tooltip"]

    @property
    def max_rank(self) -> int:
        return self._dto["maxrank"]

    @property
    def range(self) -> List[Union[int, str]]:
        return self._dto["range"]

    @property
    def cooldowns(self) -> List[float]:
        return self._dto["cooldown"]

    @property
    def costs(self) -> List[int]:
        return self._dto["cost"]

    @property
    def key(self) -> str:
        return self._dto["key"]

    @property
    def description(self) -> str:
        return self._dto["description"]

    @property
    def alternative_images(self) -> List[ImageData]:
        return [ImageData.from_dto(alt) for alt in self._dto["altimages"]]

    @property
    def name(self) -> str:
        return self._dto["name"]


class BlockItemData(CoreData):
    _renamed = {}

    @property
    def count(self) -> int:
        return self._dto["count"]

    @property
    def id(self) -> int:
        return self._dto["id"]


class BlockData(CoreData):
    _renamed = {"rec_math": "recMath"}

    @property
    def items(self) -> list:
        return [BlockItemData.from_dto(item) for item in self._dto["items"]]

    @property
    def rec_math(self) -> bool:
        return self._dto["recMath"]

    @property
    def type(self) -> str:
        return self._dto["type"]


class RecommendedData(CoreData):
    _renamed = {"item_sets": "blocks"}

    @property
    def map(self) -> str:
        return self._dto["map"]

    @property
    def item_sets(self) -> List[BlockData]:
        return [BlockData.from_dto(item_set) for item_set in self._dto["blocks"]]

    @property
    def champion(self) -> str:
        return self._dto["champion"]

    @property
    def title(self) -> str:
        return self._dto["title"]

    @property
    def priority(self) -> bool:
        return self._dto["priority"]

    @property
    def mode(self) -> str:
        return self._dto["mode"]

    @property
    def type(self) -> str:
        return self._dto["type"]


class PassiveData(CoreData):
    _renamed = {"sanitized_description": "sanitizedDescription"}

    @property
    def image(self) -> ImageData:
        return ImageData.from_dto(self._dto["image"])

    @property
    def sanitized_description(self) -> str:
        return self._dto["sanitizedDescription"]

    @property
    def name(self) -> str:
        return self._dto["name"]

    @property
    def description(self) -> str:
        return self._dto["description"]


class SkinData(CoreData):
    _renamed = {"number": "num"}

    @property
    def number(self) -> int:
        return self._dto["num"]

    @property
    def name(self) -> str:
        return self._dto["name"]

    @property
    def id(self) -> int:
        return self._dto["id"]


class StatsData(CoreData):
    _renamed = {"armor_per_level": "armorperlevel",
                "health_per_level": "hpperlevel",
                "attack_damage": "attackdamage",
                "mana_per_level": "mpperlevel",
                "attack_speed_offset": "attackspeedoffset",
                "health": "hp",
                "health_regen_per_level": "hpregenperlevel",
                "percent_attack_speed_per_level": "attackspeedperlevel",
                "attack_range": "attackrange",
                "attack_damage_per_level": "attackdamageperlevel",
                "mana_regen_per_level": "mpregenperlevel",
                "mana": "mp",
                "magic_resist_per_level": "spellblockperlevel",
                "critical_strike_chance": "crit",
                "mana_regen": "mpregen",
                "magic_resist": "spellblock",
                "health_regen": "hpregen",
                "critical_strike_chance_per_level": "critperlevel"}

    @property
    def armor_per_level(self) -> float:
        return self._dto["armorperlevel"]

    @property
    def health_per_level(self) -> float:
        return self._dto["hpperlevel"]

    @property
    def attack_damage(self) -> float:
        return self._dto["attackdamage"]

    @property
    def mana_per_level(self) -> float:
        return self._dto["mpperlevel"]

    @property
    def attack_speed_offset(self) -> float:
        return self._dto["attackspeedoffset"]

    @property
    def armor(self) -> float:
        return self._dto["armor"]

    @property
    def health(self) -> float:
        return self._dto["hp"]

    @property
    def health_regen_per_level(self) -> float:
        return self._dto["hpregenperlevel"]

    @property
    def magic_resist(self) -> float:
        return self._dto["spellblock"]

    @property
    def attack_range(self) -> float:
        return self._dto["attackrange"]

    @property
    def movespeed(self) -> float:
        return self._dto["movespeed"]

    @property
    def attack_damage_per_level(self) -> float:
        return self._dto["attackdamageperlevel"]

    @property
    def mana_regen_per_level(self) -> float:
        return self._dto["mpregenperlevel"]

    @property
    def mana(self) -> float:
        return self._dto["mp"]

    @property
    def magic_resist_per_level(self) -> float:
        return self._dto["spellblockperlevel"]

    @property
    def critical_strike_chance(self) -> float:
        return self._dto["crit"]

    @property
    def mana_regen(self) -> float:
        return self._dto["mpregen"]

    @property
    def percent_attack_speed_per_level(self) -> float:
        return self._dto["attackspeedperlevel"]

    @property
    def health_regen(self) -> float:
        return self._dto["hpregen"]

    @property
    def critical_strike_chance_per_level(self) -> float:
        return self._dto["critperlevel"]


class InfoData(CoreData):
    _renamed = {}

    @property
    def difficulty(self) -> int:
        return self._dto["difficulty"]

    @property
    def attack(self) -> int:
        return self._dto["attack"]

    @property
    def defense(self) -> int:
        return self._dto["defense"]

    @property
    def magic(self) -> int:
        return self._dto["magic"]


class ChampionData(CoreData):
    _dto_type = dto.ChampionDto
    _renamed = {"ally_tips": "allytips", "enemy_tips": "enemytips", "recommended_itemsets": "recommended", "resource": "partype", "included_data": "includedData"}

    @property
    def region(self) -> str:
        return self._dto["region"]

    @property
    def version(self) -> str:
        return self._dto["version"]

    @property
    def locale(self) -> str:
        return self._dto["locale"]

    @property
    def included_data(self) -> Set[str]:
        return self._dto["includedData"]

    @property
    def id(self) -> int:
        return self._dto["id"]

    @property
    def ally_tips(self) -> List[str]:
        return self._dto["allytips"]

    @property
    def enemy_tips(self) -> List[str]:
        return self._dto["enemytips"]

    @property
    def name(self) -> str:
        return self._dto["name"]

    @property
    def title(self) -> str:
        return self._dto["title"]

    @property
    def blurb(self) -> str:
        return self._dto["blurb"]

    @property
    def key(self) -> str:
        return self._dto["key"]

    @property
    def lore(self) -> str:
        return self._dto["lore"]

    @property
    def resource(self) -> str:
        return self._dto["partype"]

    @property
    def tags(self) -> List[str]:
        return self._dto["tags"]

    @property
    def recommended_itemsets(self) -> List[RecommendedData]:
        return [RecommendedData.from_dto(item) for item in self._dto["recommended"]]

    @property
    def info(self) -> InfoData:
        return InfoData.from_dto(self._dto["info"])

    @property
    def stats(self) -> StatsData:
        return StatsData.from_dto(self._dto["stats"])

    @property
    def image(self) -> ImageData:
        return ImageData.from_dto(self._dto["image"])

    @property
    def skins(self) -> List[SkinData]:
        return [SkinData.from_dto(skin) for skin in self._dto["skins"]]

    @property
    def passive(self) -> PassiveData:
        return PassiveData.from_dto(self._dto["passive"])

    @property
    def spells(self) -> List[ChampionSpellData]:
        return [ChampionSpellData.from_dto(spell) for spell in self._dto["spells"]]


##############
# Core Types #
##############


class Champions(CassiopeiaGhostList):
    _data_types = {ChampionListData}

    def __init__(self, *args, region: Union[Region, str] = None, version: str = None, locale: str = None, included_data: Set[str] = None):
        if region is None:
            region = settings.default_region
        if not isinstance(region, Region):
            region = Region(region)
        if included_data is None:
            included_data = {"all"}
        if locale is None:
            locale = region.default_locale
        kwargs = {"region": region, "included_data": included_data, "locale": locale}
        if version:
            kwargs["version"] = version
        super().__init__(*args, **kwargs)

    def __get_query__(self):
        return {"region": self.region, "platform": self.platform, "version": self.version, "locale": self.locale, "includedData": self.included_data}

    def __load_hook__(self, load_group: CoreData, data: CoreData) -> None:
        self.clear()
        from ...transformers.staticdata import StaticDataTransformer
        SearchableList.__init__(self, [StaticDataTransformer.champion_data_to_core(None, c) for c in data])
        super().__load_hook__(load_group, data)

    @lazy_property
    def region(self) -> Region:
        """The region for this champion."""
        return Region(self._data[ChampionListData].region)

    @lazy_property
    def platform(self) -> Platform:
        """The platform for this champion."""
        return self.region.platform

    @property
    def version(self) -> str:
        """The version for this champion."""
        try:
            return self._data[ChampionListData].version
        except KeyError:
            version = get_latest_version(region=self.region, endpoint="champion")
            self(version=version)
            return self._data[ChampionListData].version

    @property
    def locale(self) -> str:
        """The locale for this champion."""
        return self._data[ChampionListData].locale

    @property
    def included_data(self) -> Set[str]:
        """A set of tags to return additonal information for this champion when it's loaded."""
        return self._data[ChampionListData].included_data


@searchable({str: ["key"]})
class SpellVars(CassiopeiaObject):
    _data_types = {SpellVarsData}

    @property
    def ranks_with(self) -> str:
        """Well, we don't know what this one is. let us know if you figure it out."""
        return self._data[SpellVarsData].ranks_with

    @property
    def dynamic(self) -> str:
        """Well, we don't know what this one is. let us know if you figure it out."""
        return self._data[SpellVarsData].dynamic

    @property
    def link(self) -> str:
        """Stat this spell scales from."""
        return self._data[SpellVarsData].link

    @lazy_property
    def coefficients(self) -> List[float]:
        """The scaling coefficients for this spell."""
        return SearchableList(self._data[SpellVarsData].coefficients)

    @property
    def key(self) -> str:
        """Well, we don't know what this one is. let us know if you figure it out."""
        return self._data[SpellVarsData].key


@searchable({str: ["name", "key", "keywords", "resource"], Resource: ["resource"]})
class ChampionSpell(CassiopeiaObject):
    _data_types = {ChampionSpellData}

    @lazy_property
    def keywords(self) -> List[str]:
        """The keywords for this spell."""
        return SearchableList(self._data[ChampionSpellData].level_up_tips.keywords)

    @property
    def effects_by_level(self) -> List[str]:
        """The level-up changes, level-by-level."""
        return SearchableList(self._data[ChampionSpellData].level_up_tips.effects)

    @lazy_property
    def variables(self) -> List[SpellVars]:
        """Contains spell data."""
        return SearchableList(SpellVars.from_data(v) for v in self._data[ChampionSpellData].variables)

    @property
    def resource(self) -> Resource:
        """The resource consumed when using this spell."""
        return Resource(self._data[ChampionSpellData].resource)

    @lazy_property
    def image_info(self) -> Image:
        """The info about the spell's image, which can be pulled from datadragon."""
        return Image(self._data[ChampionSpellData].image)

    @property
    def sanitized_description(self) -> str:
        """The spell's sanitized description."""
        return self._data[ChampionSpellData].sanitized_description

    @property
    def sanitized_tooltip(self) -> str:
        """The spell's sanitized tooltip."""
        return self._data[ChampionSpellData].sanitized_tooltip

    @property
    def effects(self) -> List[List[float]]:
        """The level-by-level replacements for {{ e# }} tags in other values."""
        return SearchableList(self._data[ChampionSpellData].effects)

    @property
    def tooltip(self) -> str:
        """The spell's tooltip."""
        return self._data[ChampionSpellData].tooltip

    @property
    def max_rank(self) -> int:
        """The maximum rank this spell can attain."""
        return self._data[ChampionSpellData].max_rank

    @property
    def range(self) -> List[Union[int, str]]:
        """The maximum range of this spell. `self` if it has no range."""
        return SearchableList(self._data[ChampionSpellData].range)

    @property
    def cooldowns(self) -> List[float]:
        """The cooldowns of this spell (per level)."""
        return SearchableList(self._data[ChampionSpellData].cooldowns)

    @property
    def costs(self) -> List[int]:
        """The resource costs of this spell (per level)."""
        return SearchableList(self._data[ChampionSpellData].costs)

    @property
    def key(self) -> str:
        """The spell's key."""
        return self._data[ChampionSpellData].key

    @property
    def description(self) -> str:
        """The spell's description."""
        return self._data[ChampionSpellData].description

    @lazy_property
    def alternative_images(self) -> List[Image]:
        """The alternative images for this spell. These won't exist after patch NN, when Riot standardized all images."""
        return SearchableList(Image(alt) for alt in self._data[ChampionSpellData].alternative_images)

    @property
    def name(self) -> str:
        """The spell's name."""
        return self._data[ChampionSpellData].name


@searchable({str: ["type", "items"], Item: ["items"]})
class ItemSet(CassiopeiaObject):
    _data_types = {BlockData}

    @property
    def items(self) -> Dict[Item, int]:
        """A dictionary of items mapped to how many of them are recommended."""
        return SearchableDictionary({Item(id=item.id): item.count for item in self._data[BlockData].items})

    @property
    def rec_math(self) -> bool:
        """Well, we don't know what this one is. let us know if you figure it out."""
        return self._data[BlockData].rec_math

    @property
    def type(self) -> str:
        """The item set's type (e.g. starting items)."""
        return self._data[BlockData].type


@searchable({str: ["item_sets", "map", "title", "mode", "type"], Item: ["item_sets"], GameMode: ["mode"]})
class RecommendedItems(CassiopeiaObject):
    _data_types = {RecommendedData}

    @property
    def map(self) -> Map:
        """The name of the map these recommendations are for."""
        return Map(self._data[RecommendedData].map)

    @lazy_property
    def item_sets(self) -> List[ItemSet]:
        """The recommended item sets."""
        return SearchableList(ItemSet.from_data(itemset) for itemset in self._data[RecommendedData].item_sets)

    @property
    def champion(self) -> "Champion":
        """The name of the champion to whom this set of recommendations belongs."""
        return Champion(name=self._data[RecommendedData].champion)

    @property
    def title(self) -> str:
        """The title of these recommendations."""
        return self._data[RecommendedData].title

    @property
    def priority(self) -> bool:
        """Whether this is a priority recommendation."""
        return self._data[RecommendedData].priority

    @property
    def mode(self) -> GameMode:
        """The game mode these recommendations are for."""
        return GameMode(self._data[RecommendedData].mode)

    @property
    def type(self) -> str:
        """The type of recommendation."""
        return self._data[RecommendedData].type


@searchable({str: ["name"]})
class Passive(CassiopeiaObject):
    _data_types = {PassiveData}

    @lazy_property
    def image_info(self) -> Image:
        """The info about the spell's image, which can be pulled from datadragon."""
        return Image(self._data[PassiveData].image)

    @property
    def sanitized_description(self) -> str:
        """The spell's sanitized description."""
        return self._data[PassiveData].sanitized_description

    @property
    def name(self) -> str:
        """The spell's name."""
        return self._data[PassiveData].name

    @property
    def description(self) -> str:
        """The spells' description."""
        return self._data[PassiveData].description


@searchable({str: ["name", "splash_url", "loading_image_url"], int: ["id"]})
class Skin(CassiopeiaObject):
    _data_types = {SkinData}

    def __init__(self, **kwargs):
        self.__champion_key = kwargs.pop("champion_key")
        super().__init__(**kwargs)

    @property
    def champion_key(self) -> str:
        """The key for the champion this belongs to."""
        return self.__champion_key

    @property
    def number(self) -> int:
        """The skin number."""
        return self._data[SkinData].number

    @property
    def name(self) -> str:
        """The skin's name."""
        return self._data[SkinData].name

    @property
    def id(self) -> int:
        """The skin's ID."""
        return self._data[SkinData].id

    @property
    def splash_url(self) -> str:
        """The skin's splash art url."""
        return "http://ddragon.leagueoflegends.com/cdn/img/champion/splash/{}_{}.jpg".format(self.champion_key, self.number)

    @property
    def loading_image_url(self) -> str:
        """The skin's loading screen image url."""
        return "http://ddragon.leagueoflegends.com/cdn/img/champion/loading/{}_{}.jpg".format(self.champion_key, self.number)

    @lazy_property
    def splash(self) -> PILImage:
        """The skin's splash art."""
        return settings.pipeline.get(PILImage, query={"url": self.splash_url})

    @lazy_property
    def loading_image(self) -> PILImage:
        """The skin's loading screen image."""
        return settings.pipeline.get(PILImage, query={"url": self._loading_image_url})


class Stats(CassiopeiaObject):
    _data_types = {StatsData}

    @property
    def armor_per_level(self) -> float:
        return self._data[StatsData].armor_per_level

    @property
    def health_per_level(self) -> float:
        return self._data[StatsData].health_per_level

    @property
    def attack_damage(self) -> float:
        return self._data[StatsData].attack_damage

    @property
    def mana_per_level(self) -> float:
        return self._data[StatsData].mana_per_level

    @property
    def attack_speed(self) -> float:
        return 0.625 / (1.0 + self._data[StatsData].attack_speed_offset)

    @property
    def armor(self) -> float:
        return self._data[StatsData].armor

    @property
    def health(self) -> float:
        return self._data[StatsData].health

    @property
    def health_regen_per_level(self) -> float:
        return self._data[StatsData].health_regen_per_level

    @property
    def magic_resist(self) -> float:
        return self._data[StatsData].magic_resist

    @property
    def attack_range(self) -> float:
        return self._data[StatsData].attack_range

    @property
    def movespeed(self) -> float:
        return self._data[StatsData].movespeed

    @property
    def attack_damage_per_level(self) -> float:
        return self._data[StatsData].attack_damage_per_level

    @property
    def mana_regen_per_level(self) -> float:
        return self._data[StatsData].mana_regen_per_level

    @property
    def mana(self) -> float:
        return self._data[StatsData].mana

    @property
    def magic_resist_per_level(self) -> float:
        return self._data[StatsData].magic_resist_per_level

    @property
    def critical_strike_chance(self) -> float:
        return self._data[StatsData].critical_strike_chance

    @property
    def mana_regen(self) -> float:
        return self._data[StatsData].mana_regen

    @property
    def percent_attack_speed_per_level(self) -> float:
        return self._data[StatsData].percent_attack_speed_per_level / 100.0

    @property
    def health_regen(self) -> float:
        return self._data[StatsData].health_regen

    @property
    def critical_strike_chance_per_level(self) -> float:
        return self._data[StatsData].critperlevel


class Info(CassiopeiaObject):
    _data_types = {InfoData}

    @property
    def difficulty(self) -> int:
        """How Riot rates the difficulty of this champion."""
        return self._data[InfoData].difficulty

    @property
    def attack(self) -> int:
        """How attack-oriented Riot rates this champion."""
        return self._data[InfoData].attack

    @property
    def defense(self) -> int:
        """How defense-oriented Riot rates this champion."""
        return self._data[InfoData].defense

    @property
    def magic(self) -> int:
        """How magic-oriented Riot rates this champion."""
        return self._data[InfoData].magic


@searchable({str: ["name", "key", "title", "region", "platform", "locale", "resource", "tags"], int: ["id"], Region: ["region"], Platform: ["platform"], bool: ["free_to_play"], Resource: ["resource"]})
class Champion(CassiopeiaGhost):
    _data_types = {ChampionData, ChampionStatusData}
    _load_types = {ChampionData: ChampionListData, ChampionStatusData: ChampionStatusListData}

    def __init__(self, *, id: int = None, name: str = None, key: str = None, region: Union[Region, str] = None, version: str = None, locale: str = None, included_data: Set[str] = None):
        if region is None:
            region = settings.default_region
        if not isinstance(region, Region):
            region = Region(region)
        if included_data is None:
            included_data = {"all"}
        if locale is None:
            locale = region.default_locale
        kwargs = {"region": region, "included_data": included_data, "locale": locale}
        if id is not None:
            kwargs["id"] = id
        if key is not None:
            kwargs["key"] = key
        if name is not None:
            kwargs["name"] = name
        if version is not None:
            kwargs["version"] = version
        super().__init__(**kwargs)

    def __get_query__(self):
        return {"region": self.region, "platform": self.platform, "version": self.version, "locale": self.locale, "includedData": self.included_data}

    def __load_hook__(self, load_group: CoreData, data: CoreData) -> None:
        def find_matching_attribute(datalist, attrname, attrvalue):
            for item in datalist:
                if getattr(item, attrname, None) == attrvalue:
                    return item

        # The `data` is a list of champion data instances
        if "id" in self._data[ChampionData]._dto or "id" in self._data[ChampionStatusData]._dto:
            find = "id", self.id
        elif "name" in self._data[ChampionData]._dto or "name" in self._data[ChampionStatusData]._dto:
            find = "name", self.name
        else:
            raise RuntimeError("Impossible!")
        if load_group is ChampionData:
            data = find_matching_attribute(data, *find)
        else:
            data = find_matching_attribute(data, *find)
        super().__load_hook__(load_group, data)

    # What do we do about params like this that can exist in both data objects?
    # They will be set on both data objects always, so we can choose either one to return.
    @lazy_property
    def region(self) -> Region:
        """The region for this champion."""
        return Region(self._data[ChampionData].region)

    @lazy_property
    def platform(self) -> Platform:
        """The platform for this champion."""
        return self.region.platform

    @property
    def version(self) -> str:
        """The version for this champion."""
        try:
            return self._data[ChampionData].version
        except KeyError:
            version = get_latest_version(region=self.region, endpoint="champion")
            self(version=version)
            return self._data[ChampionData].version

    @property
    def locale(self) -> str:
        """The locale for this champion."""
        return self._data[ChampionData].locale

    @property
    def included_data(self) -> Set[str]:
        """A set of tags to return additonal information for this champion when it's loaded."""
        return self._data[ChampionData].included_data

    @CassiopeiaGhost.property(ChampionData)
    @ghost_load_on(KeyError)
    def id(self) -> int:
        """The champion's ID."""
        return self._data[ChampionData].id

    @CassiopeiaGhost.property(ChampionStatusData)
    @ghost_load_on(KeyError)
    def enabled(self) -> bool:
        """Whether or not the champion is currently enabled."""
        return self._data[ChampionStatusData].enabled

    @CassiopeiaGhost.property(ChampionStatusData)
    @ghost_load_on(KeyError)
    def custom_enabled(self) -> bool:
        """Whether or not the champion is currently enabled in custom games."""
        return self._data[ChampionStatusData].custom_enabled

    @CassiopeiaGhost.property(ChampionStatusData)
    @ghost_load_on(KeyError)
    def coop_ai_enabled(self) -> bool:
        """Whether or not the champion is currently enabled in coop and AI games."""
        return self._data[ChampionStatusData].coop_ai_enabled

    @CassiopeiaGhost.property(ChampionStatusData)
    @ghost_load_on(KeyError)
    def ranked_enabled(self) -> bool:
        """Whether or not the champion is currently enabled in ranked games."""
        return self._data[ChampionStatusData].ranked_enabled

    @CassiopeiaGhost.property(ChampionStatusData)
    @ghost_load_on(KeyError)
    def free_to_play(self) -> bool:
        """Whether or not the champion is currently free to play."""
        return self._data[ChampionStatusData].free_to_play

    @CassiopeiaGhost.property(ChampionData)
    @ghost_load_on(KeyError)
    def ally_tips(self) -> List[str]:
        """The tips for playing with this champion."""
        return self._data[ChampionData].ally_tips

    @CassiopeiaGhost.property(ChampionData)
    @ghost_load_on(KeyError)
    def enemy_tips(self) -> List[str]:
        """The tips for playing against this champion."""
        return self._data[ChampionData].enemy_tips

    @CassiopeiaGhost.property(ChampionData)
    @ghost_load_on(KeyError)
    def blurb(self) -> str:
        """A short blurb about this champion."""
        return self._data[ChampionData].blurb

    @CassiopeiaGhost.property(ChampionData)
    @ghost_load_on(KeyError)
    def key(self) -> str:
        """The champion's key."""
        return self._data[ChampionData].key

    @CassiopeiaGhost.property(ChampionData)
    @ghost_load_on(KeyError)
    def lore(self) -> str:
        """The champion's lore."""
        return self._data[ChampionData].lore

    @CassiopeiaGhost.property(ChampionData)
    @ghost_load_on(KeyError)
    def name(self) -> str:
        """The champion's name."""
        return self._data[ChampionData].name

    @CassiopeiaGhost.property(ChampionData)
    @ghost_load_on(KeyError)
    def resource(self) -> Resource:
        """The type of resource this champion uses."""
        return Resource(self._data[ChampionData].resource)

    @CassiopeiaGhost.property(ChampionData)
    @ghost_load_on(KeyError)
    def tags(self) -> List[str]:
        """The tags associated with this champion."""
        return self._data[ChampionData].tags

    @CassiopeiaGhost.property(ChampionData)
    @ghost_load_on(KeyError)
    def title(self) -> str:
        """The champion's title."""
        return self._data[ChampionData].title

    @CassiopeiaGhost.property(ChampionData)
    @ghost_load_on(KeyError)
    @lazy
    def info(self) -> Info:
        """Info about this champion."""
        return Info.from_data(self._data[ChampionData].info)

    @CassiopeiaGhost.property(ChampionData)
    @ghost_load_on(KeyError)
    @lazy
    def recommended_itemsets(self) -> List[RecommendedItems]:
        """The champion's recommended itemsets."""
        return SearchableList(RecommendedItems.from_data(item) for item in self._data[ChampionData].recommended_itemsets)

    @CassiopeiaGhost.property(ChampionData)
    @ghost_load_on(KeyError)
    @lazy
    def stats(self) -> Stats:
        """The champion's stats."""
        return Stats.from_data(self._data[ChampionData].stats)

    @CassiopeiaGhost.property(ChampionData)
    @ghost_load_on(KeyError)
    @lazy
    def skins(self) -> List[Skin]:
        """This champion's skins."""
        skins = []
        for skin in self._data[ChampionData].skins:
            skins.append(Skin.from_data((skin)))
            skins[-1]._Skin__champion_key = self.key
        return SearchableList(skins)

    @CassiopeiaGhost.property(ChampionData)
    @ghost_load_on(KeyError)
    @lazy
    def passive(self) -> Passive:
        """This champion's passive."""
        return Passive.from_data(self._data[ChampionData].passive)

    @CassiopeiaGhost.property(ChampionData)
    @ghost_load_on(KeyError)
    @lazy
    def spells(self) -> List[ChampionSpell]:
        """This champion's spells."""
        return SearchableList(ChampionSpell.from_data(spell) for spell in self._data[ChampionData].spells)

    @CassiopeiaGhost.property(ChampionData)
    @ghost_load_on(KeyError)
    @lazy
    def image(self) -> Image:
        """The image information for this champion."""
        image = Image(self._data[ChampionData].image)
        image(version=self.version)
        return image

    @lazy_property
    def sprite(self) -> Sprite:
        return self.image.sprite_info
